"""adding referer to authenticationrequest

Revision ID: 3a47ce736a37
Revises: 5abb65a8cb91
Create Date: 2023-07-18 03:38:18.238072

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "3a47ce736a37"
down_revision = "5abb65a8cb91"
branch_labels = None
depends_on = None


def upgrade():
    op.add_column(
        "authenticationrequest", sa.Column("referer", sa.String(), nullable=True)
    )


def downgrade():
    op.drop_column("authenticationrequest", "referer")



"""add_notice_key_to_notices

Revision ID: 8a71872089e4
Revises: 2661f31daffb
Create Date: 2023-05-18 19:48:33.268790

"""

import sqlalchemy as sa
from alembic import op
from fideslang.validation import FidesKey, FidesValidationError, validate_fides_key

# revision identifiers, used by Alembic.
from sqlalchemy import text
from sqlalchemy.engine import ResultProxy

revision = "8a71872089e4"
down_revision = "2661f31daffb"
branch_labels = None
depends_on = None


def validate_fides_key_suitability(names: ResultProxy, table_name: str) -> None:
    for row in names:
        name: str = row["name"].strip(" ").replace(" ", "_")
        try:
            validate_fides_key(name)
        except FidesValidationError as exc:
            raise Exception(
                f"Cannot auto-migrate, adjust existing {table_name} name: '{name}' to remove invalid characters: {exc}."
            )


def upgrade():
    """Add new non-nullable notice_key fields to privacy notice and notice history tables
    and automatically create existing notice keys from the notice names.

    Notice keys are not unique.
    """
    bind = op.get_bind()
    existing_history_names: ResultProxy = bind.execute(
        text("select name from privacynoticehistory;")
    )
    validate_fides_key_suitability(existing_history_names, "privacynoticehistory")

    existing_notice_names: ResultProxy = bind.execute(
        text("select name from privacynotice;")
    )
    validate_fides_key_suitability(existing_notice_names, "privacynotice")

    op.add_column("privacynotice", sa.Column("notice_key", sa.String(), nullable=True))
    op.add_column(
        "privacynoticehistory", sa.Column("notice_key", sa.String(), nullable=True)
    )

    op.execute(
        "update privacynoticehistory set notice_key = LOWER(REPLACE(TRIM(name), ' ', '_'));"
    )
    op.execute(
        "update privacynotice set notice_key = LOWER(REPLACE(TRIM(name), ' ', '_'));"
    )

    op.alter_column("privacynotice", "notice_key", nullable=False)
    op.alter_column("privacynoticehistory", "notice_key", nullable=False)


def downgrade():
    op.drop_column("privacynoticehistory", "notice_key")
    op.drop_column("privacynotice", "notice_key")



"""Add plus_system_scans table

Revision ID: 6b9885e68cbb
Revises: fb6b0150d6e4
Create Date: 2022-10-24 23:36:51.166480

"""

from alembic.op import create_table, drop_table
from sqlalchemy import JSON, Boolean, Column, DateTime, Integer, String, text

# revision identifiers, used by Alembic.
revision = "6b9885e68cbb"
down_revision = "fb6b0150d6e4"
branch_labels = None
depends_on = None


def upgrade() -> None:
    create_table(
        "plus_system_scans",
        Column("id", Integer(), primary_key=True),
        Column("created_at", DateTime(timezone=True), server_default=text("now()")),
        Column("error", String(), nullable=True),
        Column("is_classified", Boolean(), default=False, nullable=False),
        Column("result", JSON(), nullable=True),
        Column("status", String(), nullable=False),
        Column("system_count", Integer(), autoincrement=False, nullable=True),
        Column("updated_at", DateTime(timezone=True), server_default=text("now()")),
    )


def downgrade() -> None:
    drop_table("plus_system_scans")



from typing import Any, Dict, Union

from botocore.exceptions import ClientError

from fides.api.schemas.storage.storage import (
    SUPPORTED_STORAGE_SECRETS,
    AWSAuthMethod,
    StorageSecrets,
    StorageType,
)
from fides.api.util.aws_util import get_aws_session


def secrets_are_valid(
    secrets: SUPPORTED_STORAGE_SECRETS,
    storage_type: Union[StorageType, str],
) -> bool:
    """Authenticates upload destination with appropriate upload method"""
    if not isinstance(storage_type, StorageType):
        # try to coerce into an enum
        try:
            storage_type = StorageType[storage_type]
        except KeyError:
            raise ValueError(
                "storage_type argument must be a valid StorageType enum member."
            )
    uploader: Any = _get_authenticator_from_config(storage_type)
    return uploader(secrets)


def _s3_authenticator(secrets: Dict[StorageSecrets, Any]) -> bool:
    """Authenticates secrets for s3, returns true if secrets are valid"""
    try:
        get_aws_session(AWSAuthMethod.SECRET_KEYS.value, secrets.model_dump(mode="json"))  # type: ignore
        return True
    except ClientError:
        return False


def _get_authenticator_from_config(storage_type: StorageType) -> Any:
    """Determines which uploader method to use based on storage type"""
    return {
        StorageType.s3.value: _s3_authenticator,
    }[storage_type.value]



from contextlib import contextmanager
from typing import Generator

from fastapi import Depends
from sqlalchemy.orm import Session

from fides.api.common_exceptions import RedisNotConfigured
from fides.api.db.session import get_db_engine, get_db_session
from fides.api.util.cache import get_cache as get_redis_connection
from fides.config import CONFIG, FidesConfig
from fides.config import get_config as get_app_config
from fides.config.config_proxy import ConfigProxy

_engine = None


def get_config() -> FidesConfig:
    """Returns the config for use in dependency injection."""
    return get_app_config()


def get_db() -> Generator:
    """Return our database session"""
    try:
        db = get_api_session()
        yield db
    finally:
        db.close()


@contextmanager
def get_db_contextmanager() -> Generator[Session, None, None]:
    """Return our database session as a context manager"""
    try:
        db = get_api_session()
        yield db
    finally:
        db.close()


def get_api_session() -> Session:
    """Gets the shared database session to use for API functionality"""
    global _engine  # pylint: disable=W0603
    if not _engine:
        _engine = get_db_engine(
            config=CONFIG,
            pool_size=CONFIG.database.api_engine_pool_size,
            max_overflow=CONFIG.database.api_engine_max_overflow,
            keepalives_idle=CONFIG.database.api_engine_keepalives_idle,
            keepalives_interval=CONFIG.database.api_engine_keepalives_interval,
            keepalives_count=CONFIG.database.api_engine_keepalives_count,
        )
    SessionLocal = get_db_session(CONFIG, engine=_engine)
    db = SessionLocal()
    return db


def get_config_proxy(db: Session = Depends(get_db)) -> ConfigProxy:
    return ConfigProxy(db)


def get_cache() -> Generator:
    """Return a connection to our Redis cache"""
    if not CONFIG.redis.enabled:
        raise RedisNotConfigured(
            "Application redis cache required, but it is currently disabled! Please update your application configuration to enable integration with a Redis cache."
        )
    yield get_redis_connection()



from typing import List

from sqlalchemy.engine import LegacyCursorResult  # type: ignore

from fides.api.schemas.connection_configuration.connection_secrets_mariadb import (
    MariaDBSchema,
)
from fides.api.service.connectors.sql_connector import SQLConnector
from fides.api.util.collection_util import Row
from fides.config import get_config

CONFIG = get_config()


class MariaDBConnector(SQLConnector):
    """Connector specific to MariaDB"""

    secrets_schema = MariaDBSchema

    def build_uri(self) -> str:
        """Build URI of format mariadb+pymysql://[user[:password]@][netloc][:port][/dbname]"""
        config = self.secrets_schema(**self.configuration.secrets or {})

        user_password = ""
        if config.username:
            user = config.username
            password = f":{config.password}" if config.password else ""
            user_password = f"{user}{password}@"

        netloc = config.host
        port = f":{config.port}" if config.port else ""
        dbname = f"/{config.dbname}" if config.dbname else ""
        url = f"mariadb+pymysql://{user_password}{netloc}{port}{dbname}"
        return url

    @staticmethod
    def cursor_result_to_rows(results: LegacyCursorResult) -> List[Row]:
        """
        Convert SQLAlchemy results to a list of dictionaries
        """
        return SQLConnector.default_cursor_result_to_rows(results)



"""remove cascade delete from property linking table

Revision ID: 7799269e7078
Revises: 2e9aba76c322
Create Date: 2024-03-14 05:53:23.282985

"""

from alembic import op

# revision identifiers, used by Alembic.
revision = "7799269e7078"
down_revision = "2e9aba76c322"
branch_labels = None
depends_on = None


def upgrade():
    op.drop_constraint(
        "plus_privacy_experience_config_property_property_id_fkey",
        "plus_privacy_experience_config_property",
        type_="foreignkey",
    )
    op.drop_constraint(
        "plus_privacy_experience_confi_privacy_experience_config_id_fkey",
        "plus_privacy_experience_config_property",
        type_="foreignkey",
    )
    op.create_foreign_key(
        "plus_privacy_experience_config_property_property_id_fkey",
        "plus_privacy_experience_config_property",
        "plus_property",
        ["property_id"],
        ["id"],
    )
    op.create_foreign_key(
        "plus_privacy_experience_confi_privacy_experience_config_id_fkey",
        "plus_privacy_experience_config_property",
        "privacyexperienceconfig",
        ["privacy_experience_config_id"],
        ["id"],
    )


def downgrade():
    op.drop_constraint(
        "plus_privacy_experience_config_property_property_id_fkey",
        "plus_privacy_experience_config_property",
        type_="foreignkey",
    )
    op.drop_constraint(
        "plus_privacy_experience_confi_privacy_experience_config_id_fkey",
        "plus_privacy_experience_config_property",
        type_="foreignkey",
    )
    op.create_foreign_key(
        "plus_privacy_experience_confi_privacy_experience_config_id_fkey",
        "plus_privacy_experience_config_property",
        "privacyexperienceconfig",
        ["privacy_experience_config_id"],
        ["id"],
        ondelete="CASCADE",
    )
    op.create_foreign_key(
        "plus_privacy_experience_config_property_property_id_fkey",
        "plus_privacy_experience_config_property",
        "plus_property",
        ["property_id"],
        ["id"],
        ondelete="CASCADE",
    )



# pylint: disable=useless-import-alias

# Because these modules are imported into the __init__.py and used elsewhere they need
# to be explicitly exported in order to prevent implicit reexport errors in mypy.
# This is done by importing "as": `from fides.module import function as function`.

from .get_email_template import get_email_template as get_email_template



from __future__ import annotations

from enum import Enum
from typing import Any, Dict, List, Optional, Set, Type

from sqlalchemy import Boolean, Column
from sqlalchemy import Enum as EnumColumn
from sqlalchemy import Float, ForeignKey, String, UniqueConstraint
from sqlalchemy.dialects.postgresql import ARRAY, JSONB
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.orm import Query, RelationshipProperty, Session, relationship
from sqlalchemy.orm.dynamic import AppenderQuery

from fides.api.db.base_class import Base
from fides.api.models import (
    create_historical_data_from_record,
    dry_update_data,
    update_if_modified,
)
from fides.api.models.location_regulation_selections import PrivacyNoticeRegion
from fides.api.models.privacy_notice import PrivacyNotice
from fides.api.models.property import Property
from fides.api.schemas.language import SupportedLanguage


class ComponentType(Enum):
    """
    The component type - not formalized in the db

    Overlay type has been deprecated but can't be removed for backwards compatibility
    without significant data migrations.
    """

    overlay = "overlay"  # Deprecated. DO NOT REMOVE.
    banner_and_modal = "banner_and_modal"
    modal = "modal"
    privacy_center = "privacy_center"
    tcf_overlay = "tcf_overlay"  # TCF Banner + modal combined
    headless = "headless"


class Layer1ButtonOption(Enum):
    """
    Layer 1 button options - not formalized in the db
    """

    ACKNOWLEDGE = "acknowledge"
    OPT_IN_OPT_OUT = "opt_in_opt_out"


# Fides JS UX Types - there should only be one of these defined per region
FidesJSUXTypes: List[ComponentType] = [
    ComponentType.banner_and_modal,
    ComponentType.modal,
    ComponentType.headless,
]

# Fides JS Overlay Types - there should only be one of these defined per region + property
FidesJSOverlayTypes: List[ComponentType] = [
    ComponentType.banner_and_modal,
    ComponentType.modal,
    ComponentType.tcf_overlay,
    ComponentType.headless,
]


class PrivacyExperienceConfigBase:
    """
    Common fields shared between:
    - ExperienceConfigTemplate
    - PrivacyExperienceConfig
    - PrivacyExperienceConfigHistory

    These are non-translated fields.
    """

    allow_language_selection = Column(
        Boolean,
    )
    auto_detect_language = Column(
        Boolean,
    )
    auto_subdomain_cookie_deletion = Column(
        Boolean,
        nullable=True,
        default=True,
    )  # base is nullable for privacy experience config history

    disabled = Column(Boolean, nullable=False, default=True)

    dismissable = Column(Boolean)

    show_layer1_notices = Column(Boolean, nullable=True, default=False)

    @declared_attr
    def layer1_button_options(cls) -> Column:
        return Column(
            EnumColumn(Layer1ButtonOption),
            nullable=True,
            index=False,
        )

    @declared_attr
    def component(cls) -> Column:
        return Column(
            EnumColumn(ComponentType),
            nullable=False,
            index=True,
        )

    name = Column(String)


class ExperienceConfigTemplate(PrivacyExperienceConfigBase, Base):
    """Table for out-of-the-box Experience Configurations"""

    allow_language_selection = Column(
        Boolean, nullable=False, default=False, server_default="f"
    )  # Overrides PrivacyExperienceConfigBase to make non-nullable
    auto_detect_language = Column(
        Boolean, nullable=False, default=True, server_default="t"
    )  # Overrides PrivacyExperienceConfigBase to make non-nullable
    auto_subdomain_cookie_deletion = Column(
        Boolean, nullable=False, default=True, server_default="t"
    )  # Overrides PrivacyExperienceConfigBase to make non-nullable
    dismissable = Column(
        Boolean, nullable=False, default=True, server_default="t"
    )  # Overrides PrivacyExperienceConfigBase to make non-nullable
    name = Column(
        String, nullable=False
    )  # Overriding PrivacyExperienceConfigBase to make non-nullable

    privacy_notice_keys = Column(
        ARRAY(String)
    )  # A list of notice keys which should correspond to a subset of out-of-the-box notice keys
    regions = Column(
        ARRAY(EnumColumn(PrivacyNoticeRegion, native_enum=False)),
    )
    translations = Column(
        ARRAY(JSONB)
    )  # A list of all available out of the box translations


class ExperienceTranslationBase:
    """Base schema for fields shared between ExperienceTranslation and PrivacyExperienceConfigHistory.

    These are translated fields
    """

    accept_button_label = Column(String)
    acknowledge_button_label = Column(String)
    banner_description = Column(String)
    banner_title = Column(String)
    description = Column(String)
    is_default = Column(Boolean, nullable=False, default=False)

    language = Column(
        EnumColumn(
            SupportedLanguage,
            native_enum=False,
            values_callable=lambda x: [
                i.value for i in x
            ],  # allows enum _values_ to be stored rather than name
        ),
    )

    privacy_policy_link_label = Column(String)
    privacy_policy_url = Column(String)
    privacy_preferences_link_label = Column(String)
    purpose_header = Column(String)
    modal_link_label = Column(String)
    reject_button_label = Column(String)
    save_button_label = Column(String)
    title = Column(String)


class PrivacyExperienceConfig(PrivacyExperienceConfigBase, Base):
    """
    The Privacy Experience Configuration model that stores shared configuration for Privacy Experiences.

    - Translations, Notices, and Regions (via Privacy Experiences) are linked to this resource.
    """

    allow_language_selection = Column(
        Boolean, nullable=False, default=False, server_default="f"
    )  # Overrides PrivacyExperienceConfigBase to make non-nullable
    auto_detect_language = Column(
        Boolean, nullable=False, default=True, server_default="t"
    )  # Overrides PrivacyExperienceConfigBase to make non-nullable
    auto_subdomain_cookie_deletion = Column(
        Boolean, nullable=False, default=True, server_default="t"
    )  # Overrides PrivacyExperienceConfigBase to make non-nullable
    disabled = Column(
        Boolean, nullable=False, default=True, index=True
    )  # Overridding PrivacyExperienceConfigBase to index
    dismissable = Column(
        Boolean, nullable=False, default=True, server_default="t"
    )  # Overrides PrivacyExperienceConfigBase to make non-nullable
    name = Column(
        String, nullable=False
    )  # Overriding PrivacyExperienceConfigBase to make non-nullable
    origin = Column(
        String, ForeignKey(ExperienceConfigTemplate.id_field_path)
    )  # The template from which this config was created if applicable

    # Relationships
    experiences = relationship(
        "PrivacyExperience",
        back_populates="experience_config",
        lazy="dynamic",
        cascade="all,delete",
    )

    privacy_notices: RelationshipProperty[List[PrivacyNotice]] = relationship(
        "PrivacyNotice",
        secondary="experiencenotices",
        backref="experience_configs",
        lazy="selectin",
    )

    translations: RelationshipProperty[List[ExperienceTranslation]] = relationship(
        "ExperienceTranslation",
        backref="privacy_experience_config",
        lazy="selectin",
        order_by="ExperienceTranslation.created_at",
    )

    properties: RelationshipProperty[List[Property]] = relationship(
        "Property",
        secondary="plus_privacy_experience_config_property",
        back_populates="experiences",
        lazy="selectin",
    )

    @property
    def regions(self) -> List[PrivacyNoticeRegion]:
        """Return the regions using this experience config"""
        return [exp.region for exp in self.experiences]  # type: ignore[attr-defined]

    def get_translation_by_language(
        self, db: Session, language: Optional[SupportedLanguage]
    ) -> Optional[ExperienceTranslation]:
        """Lookup a translation on an ExperienceConfig by language if it exists"""
        if not language:
            return None
        return (
            db.query(ExperienceTranslation)
            .filter(
                ExperienceTranslation.language == language,
                ExperienceTranslation.experience_config_id == self.id,
            )
            .first()
        )

    @classmethod
    def create(
        cls: Type[PrivacyExperienceConfig],
        db: Session,
        *,
        data: dict[str, Any],
        check_name: bool = False,
    ) -> PrivacyExperienceConfig:
        """
        Creates an Experience Config which is a large exercise!

        - Creates the ExperienceConfig
        - For each Translation supplied, creates an ExperienceTranslation and a historical
        record combining details from the translation and the Experience Config
        - Links regions to the ExperienceConfig (by creating/updating associated PrivacyExperiences
        and adding a FK back to the config)
        - Adds/removes Notices to/from the Experience Config
        """
        translations = data.pop("translations", [])
        regions = data.pop("regions", [])
        privacy_notice_ids = data.pop("privacy_notice_ids", [])
        properties = data.pop("properties", [])
        data.pop(
            "id", None
        )  # Default templates have ids but we don't want to use them here

        experience_config: PrivacyExperienceConfig = super().create(
            db=db, data=data, check_name=check_name
        )

        for translation_data in translations:
            # Create an ExperienceTranslation
            translation = ExperienceTranslation.create(
                db,
                data={**translation_data, "experience_config_id": experience_config.id},
            )
            # Version the ExperienceTranslation and the original PrivacyExperienceConfig together.
            PrivacyExperienceConfigHistory.create(
                db,
                data={**data, **translation_data, "translation_id": translation.id},
                check_name=False,
            )

        # Link regions to this PrivacyExperienceConfig via the PrivacyExperience table
        upsert_privacy_experiences_after_config_update(db, experience_config, regions)
        # Link Privacy Notices to this Privacy Experience config via the ExperienceNotices table
        link_notices_to_experience_config(
            db, notice_ids=privacy_notice_ids, experience_config=experience_config
        )
        # Link Properties to this Privacy Experience config via the PrivacyExperienceConfigProperty table
        link_properties_to_experience_config(db, properties, experience_config)

        return experience_config

    def update(self, db: Session, *, data: dict[str, Any]) -> PrivacyExperienceConfig:
        """
        Updates a PrivacyExperienceConfig and related resources which is a large exercise!

        - Updates the PrivacyExperienceConfig details
        - For each supplied translation, add, update, or delete translations so they match
        the translations in the update request
        - For each remaining translation, if the translation or the config has changed,
        create a historical record for auditing purposes
        - Link or unlink regions (via Privacy Experiences)
        - Link or unlink Privacy Notices
        """
        request_translations = data.pop("translations", [])
        regions = data.pop("regions", [])
        privacy_notice_ids = data.pop("privacy_notice_ids", [])
        properties = data.pop("properties", [])

        # Do a patch update of the existing privacy experience config if applicable
        config_updated = update_if_modified(self, db=db, data=data)

        for translation_data in request_translations:
            existing_translation: Optional[ExperienceTranslation] = (
                self.get_translation_by_language(db, translation_data.get("language"))
            )
            if existing_translation:
                # Do a patch update of the existing experience translation if applicable
                translation_updated: bool = update_if_modified(
                    existing_translation,
                    db=db,
                    data={**translation_data, "experience_config_id": self.id},
                )
                translation = existing_translation
            else:
                translation_updated = True
                # Create a new Experience Translation since one doesn't exist
                translation = ExperienceTranslation.create(
                    db,
                    data={**translation_data, "experience_config_id": self.id},
                )

            if config_updated or translation_updated:
                create_historical_record_for_config_and_translation(
                    db,
                    privacy_experience_config=self,
                    experience_translation=translation,
                )

        # Deletes any Experience Translations that were not supplied in the request
        delete_experience_config_translations(
            db=db,
            privacy_experience_config=self,
            request_translations=request_translations,
        )

        # Link regions to this PrivacyExperienceConfig via the PrivacyExperience table
        upsert_privacy_experiences_after_config_update(db, self, regions)

        # Link Privacy Notices to this Privacy Experience config via the ExperienceNotices table
        link_notices_to_experience_config(
            db, notice_ids=privacy_notice_ids, experience_config=self
        )
        # Link Properties to this Privacy Experience config via the PrivacyExperienceConfigProperty table
        link_properties_to_experience_config(db, properties, self)

        return self  # type: ignore[return-value]

    def dry_update(self, *, data: dict[str, Any]) -> PrivacyExperienceConfig:
        """
        A utility method to get an updated ExperienceConfig without saving it to the db.

        This is used to see what an object update would look like, in memory,
        without actually persisting the update to the db
        """
        updated_attributes = dry_update_data(resource=self, data_updates=data)
        updated_attributes.pop("regions", [])
        updated_attributes.pop("translations", [])
        # Updated privacy notice ids from the request will be in this format
        updated_attributes.pop("privacy_notice_ids", [])
        # Existing privacy notices on the ExperienceConfig need to be popped off here as well
        # to prevent the ExperienceConfig "dry_update" from being added to Session.new
        # (which would cause another PrivacyExperienceConfig to be created!)
        updated_attributes.pop("privacy_notices", [])
        updated_attributes.pop("properties", [])

        return PrivacyExperienceConfig(**updated_attributes)

    def dry_update_translations(
        self, data: list[dict[str, Any]]
    ) -> List[ExperienceTranslation]:
        """A utility method to get updated ExperienceTranslations without saving them to the db"""
        dry_updated_translations = []
        for translation_data in data:
            existing_translation = self.get_translation_by_language(
                Session.object_session(self), translation_data.get("language")
            )
            if existing_translation:
                updated_attributes: Dict = dry_update_data(
                    existing_translation, translation_data
                )
            else:
                updated_attributes = translation_data
            dry_updated_translations.append(ExperienceTranslation(**updated_attributes))
        return dry_updated_translations


class ExperienceTranslation(ExperienceTranslationBase, Base):
    """Stores all the translations for a given Experience Config"""

    experience_config_id = Column(
        String,
        ForeignKey(PrivacyExperienceConfig.id_field_path),
        nullable=False,
        index=True,
    )

    language = Column(
        EnumColumn(
            SupportedLanguage,
            native_enum=False,
            values_callable=lambda x: [
                i.value for i in x
            ],  # allows enum _values_ to be stored rather than name
        ),
        nullable=False,
    )  # Overridding language on ExperienceTranslationBase to make this non-nullable

    __table_args__ = (
        UniqueConstraint(
            "language", "experience_config_id", name="experience_translation"
        ),
    )

    histories: RelationshipProperty[AppenderQuery] = relationship(
        "PrivacyExperienceConfigHistory",
        backref="experience_translation",
        lazy="dynamic",
        order_by="PrivacyExperienceConfigHistory.created_at",
    )

    @property
    def version(self) -> Optional[float]:
        """Convenience property that returns the latest version number of the translation"""
        return (
            self.experience_config_history.version  # type: ignore[return-value]
            if self.experience_config_history
            else None
        )

    @property
    def experience_config_history(self) -> Optional[PrivacyExperienceConfigHistory]:
        """Convenience property that returns the experience config history for the latest version.

        Note that there are possibly many historical records for the given experience config translation,
        this just returns the current corresponding historical record.
        """
        # Histories are sorted at the relationship level
        return self.histories[-1] if self.histories.count() else None

    @property
    def privacy_experience_config_history_id(self) -> Optional[str]:
        """Convenience property that returns the experience config history id for the latest version.

        Note that there are possibly many historical records for the given experience config translation, this just returns the current
        corresponding historical record.
        """
        return (
            self.experience_config_history.id
            if self.experience_config_history
            else None
        )


class DeprecatedPrivacyExperienceConfigHistoryFields:
    """Fields we no longer collect and save, but are retaining on early records for auditing purposes"""

    banner_enabled = Column(
        String(), index=True
    )  # No longer collected, this is now a function of the experience config type itself


class PrivacyExperienceConfigHistory(
    ExperienceTranslationBase,
    PrivacyExperienceConfigBase,
    DeprecatedPrivacyExperienceConfigHistoryFields,
    Base,
):
    """Experience Config History table for auditing purposes.

    When an Experience Config and/or an Experience Translation is modified, a new version is
    created here that stores a snapshot of these records combined.  Consent reporting can contain
    an id to this resource which preserves the details of the Experience viewed by the end user.
    """

    origin = Column(String, ForeignKey(ExperienceConfigTemplate.id_field_path))

    translation_id = Column(
        String,
        ForeignKey(ExperienceTranslation.id_field_path, ondelete="SET NULL"),
        index=True,
    )  # If a translation is deleted, this is set to null, but the overall record remains in the database for reporting purposes

    version = Column(Float, nullable=False, default=1.0)


class PrivacyExperience(Base):
    """Privacy Experiences connect a region to a shared PrivacyExperienceConfig.

    Privacy Experiences are queried by end-users in a specific region.  The configuration is loaded
    from their PrivacyExperienceConfig.
    """

    region = Column(EnumColumn(PrivacyNoticeRegion), nullable=False, index=True)

    experience_config_id = Column(
        String,
        ForeignKey(PrivacyExperienceConfig.id_field_path),
        nullable=False,
        index=True,
    )

    experience_config = relationship(
        "PrivacyExperienceConfig",
        back_populates="experiences",
        uselist=False,
    )

    @property
    def component(self) -> Optional[ComponentType]:
        """For backwards compatibility, returns the component from the attached Privacy Experience Config
        This used to be stored at the PrivacyExperience level too.
        """
        if not self.experience_config_id:
            return None
        return self.experience_config.component  # type: ignore[return-value]

    @property
    def show_banner(self) -> bool:
        """Backwards compatible property for whether the banner should be shown.

        Logic used to be much more complex and calculated at runtime. Now this can be derived from just the component type
        """
        if self.component in [
            ComponentType.tcf_overlay,
            ComponentType.banner_and_modal,
            ComponentType.overlay,
        ]:  # Overlay type is deprecated. For backwards compatibility.
            return True

        return False

    # Attribute that can be added as the result of "get_related_privacy_notices". Privacy notices aren't directly
    # related to experiences.
    privacy_notices: List[PrivacyNotice] = []
    # TCF attributes that can be added at runtime as the result of "get_related_tcf_contents"
    tcf_purpose_consents: List = []
    tcf_purpose_legitimate_interests: List = []
    tcf_special_purposes: List = []
    tcf_vendor_consents: List = []
    tcf_vendor_legitimate_interests: List = []
    tcf_features: List = []
    tcf_special_features: List = []
    tcf_system_consents: List = []
    tcf_system_legitimate_interests: List = []
    gvl: Optional[Dict] = {}
    # TCF Developer-Friendly Meta added at runtime as the result of build_tc_data_for_mobile
    meta: Dict = {}

    @property
    def region_country(self) -> str:
        """The experience's country, based on naming convention of its region string."""
        return region_country(self.region.value)  # type: ignore[attr-defined]

    @staticmethod
    def get_experiences_by_region_and_component(
        db: Session, region: str, component: ComponentType
    ) -> Query:
        """Utility method to load experiences for a given region and component type"""
        return (
            db.query(PrivacyExperience)
            .join(
                PrivacyExperienceConfig,
                PrivacyExperienceConfig.id == PrivacyExperience.experience_config_id,
            )
            .filter(
                PrivacyExperience.region == region,
                PrivacyExperienceConfig.component  # pylint: disable=comparison-with-callable
                == component,
            )
        )


def upsert_privacy_experiences_after_config_update(
    db: Session,
    experience_config: PrivacyExperienceConfig,
    regions: List[PrivacyNoticeRegion],
) -> List[PrivacyNoticeRegion]:
    """
    Links regions to a PrivacyExperienceConfig by adding or removing PrivacyExperience records.
    """
    current_regions: List[PrivacyNoticeRegion] = experience_config.regions
    removed_regions: List[PrivacyNoticeRegion] = (
        [  # Regions that were not in the request, but currently attached to the Config
            PrivacyNoticeRegion(reg)
            for reg in {reg.value for reg in current_regions}.difference(
                {reg.value for reg in regions}
            )
        ]
    )

    # Delete any PrivacyExperiences whose regions are not in the request
    experience_config.experiences.filter(  # type: ignore[call-arg]
        PrivacyExperience.region.in_(removed_regions)
    ).delete()

    for region in regions:
        existing_experience = (
            db.query(PrivacyExperience)
            .filter(
                PrivacyExperience.experience_config_id == experience_config.id,
                PrivacyExperience.region == region,
            )
            .first()
        )

        if not existing_experience:
            # Create a Privacy Experience for any new regions
            PrivacyExperience.create(
                db,
                data={
                    "region": region,
                    "experience_config_id": experience_config.id,
                },
            )

    return experience_config.regions


def region_country(region: str) -> str:
    """
    Utility function to extract the country string from a region string,
    based on naming convention (i.e. `country_subregion`)
    """
    return region.split("_")[0]


def link_notices_to_experience_config(
    db: Session,
    notice_ids: List[str],
    experience_config: PrivacyExperienceConfig,
) -> List[PrivacyNotice]:
    """
    Link supplied Notices to ExperienceConfig and unlink any notices not supplied.
    """
    new_notices: Query = db.query(PrivacyNotice).filter(
        PrivacyNotice.id.in_(notice_ids)
    )

    for notice in new_notices:
        if notice not in experience_config.privacy_notices:
            experience_config.privacy_notices.append(notice)

    to_remove: Set[PrivacyNotice] = set(
        notice for notice in experience_config.privacy_notices
    ) - set(notice for notice in new_notices)

    for privacy_notice in to_remove:
        experience_config.privacy_notices.remove(privacy_notice)

    experience_config.save(db)
    return experience_config.privacy_notices


def link_properties_to_experience_config(
    db: Session,
    properties: List[Dict[str, Any]],
    experience_config: PrivacyExperienceConfig,
) -> List[Property]:
    """
    Link supplied properties to ExperienceConfig and unlink any properties not supplied.
    """
    new_properties = (
        db.query(Property)
        .filter(Property.id.in_([prop["id"] for prop in properties]))
        .all()
    )
    experience_config.properties = new_properties
    experience_config.save(db)
    return experience_config.properties


def create_historical_record_for_config_and_translation(
    db: Session,
    privacy_experience_config: PrivacyExperienceConfig,
    experience_translation: ExperienceTranslation,
) -> None:
    """
    Create a PrivacyExperienceConfigHistory record that preserves changes from the ExperienceTranslation and/or its
    PrivacyExperienceConfig for consent reporting purposes.

    The id of this record is used to save more context around what an end user was shown when saving privacy preferences.
    """
    existing_version: float = experience_translation.version or 0.0
    history_data: dict = create_historical_data_from_record(privacy_experience_config)
    history_data.pop("privacy_notices", None)
    history_data.pop("translations", None)
    history_data.pop("properties", None)

    updated_translation_data: dict = create_historical_data_from_record(
        experience_translation
    )
    # Translations have FK's back to experience config but the historical data does not
    updated_translation_data.pop("experience_config_id", None)

    # Create a historical record for reporting purposes, which versions
    # elements from both the PrivacyExperienceConfig and ExperienceTranslation
    PrivacyExperienceConfigHistory.create(
        db,
        data={
            **history_data,
            **updated_translation_data,
            "translation_id": experience_translation.id,
            "version": existing_version + 1.0,
        },
        check_name=False,
    )


def delete_experience_config_translations(
    db: Session,
    privacy_experience_config: PrivacyExperienceConfig,
    request_translations: Dict,
) -> None:
    """Removes any translations that are currently stored on the PrivacyExperienceConfig
    but not in the update request"""
    experience_translations: List[ExperienceTranslation] = (
        privacy_experience_config.translations
    )
    translations_to_remove: Set[SupportedLanguage] = set(  # type: ignore[assignment]
        translation.language for translation in experience_translations
    ).difference(
        set(translation.get("language") for translation in request_translations)
    )

    db.query(ExperienceTranslation).filter(
        ExperienceTranslation.language.in_(translations_to_remove),
        ExperienceTranslation.experience_config_id == privacy_experience_config.id,
    ).delete()
    db.commit()



from enum import Enum
from typing import Any, Dict, Optional, Set

from pydantic import Field

from fides.api.schemas.base_class import FidesSchema


class ReportType(str, Enum):
    """Enum for custom report types."""

    datamap = "datamap"
    privacy_request = "privacy_request"


class ColumnMapItem(FidesSchema):
    """A map between column keys and custom labels."""

    label: Optional[str] = Field(
        default=None, description="The custom label for the column"
    )
    enabled: Optional[bool] = Field(
        default=True, description="Whether the column is shown"
    )


class CustomReportConfig(FidesSchema):
    """The configuration for a custom report."""

    table_state: Dict[str, Any] = Field(
        default_factory=dict,
        description="Flexible dictionary storing UI-specific table state data without a fixed schema",
    )
    column_map: Dict[str, ColumnMapItem] = Field(
        default_factory=dict, description="A map between column keys and custom labels"
    )

    @property
    def columns_to_skip(self) -> Set[str]:
        return {
            key
            for key, value in self.column_map.items()  # pylint: disable=no-member
            if value.enabled is False
        }

    @property
    def custom_column_labels(self) -> Dict[str, str]:
        return {
            key: value.label
            for key, value in self.column_map.items()  # pylint: disable=no-member
            if value.label
        }



# pylint: disable=W0611
# Import all the models, so that Base has them before being
# imported by Alembic
from fides.api.db.base_class import Base
from fides.api.models.application_config import ApplicationConfig
from fides.api.models.asset import Asset
from fides.api.models.attachment import Attachment, AttachmentReference
from fides.api.models.audit_log import AuditLog
from fides.api.models.authentication_request import AuthenticationRequest
from fides.api.models.client import ClientDetail
from fides.api.models.connectionconfig import ConnectionConfig
from fides.api.models.consent_automation import ConsentAutomation
from fides.api.models.custom_asset import CustomAsset
from fides.api.models.custom_connector_template import CustomConnectorTemplate
from fides.api.models.custom_report import CustomReport
from fides.api.models.datasetconfig import DatasetConfig
from fides.api.models.db_cache import DBCache
from fides.api.models.detection_discovery import MonitorConfig, StagedResource
from fides.api.models.experience_notices import ExperienceNotices
from fides.api.models.fides_cloud import FidesCloud
from fides.api.models.fides_user import FidesUser
from fides.api.models.fides_user_invite import FidesUserInvite
from fides.api.models.fides_user_permissions import FidesUserPermissions
from fides.api.models.identity_salt import IdentitySalt
from fides.api.models.location_regulation_selections import LocationRegulationSelections
from fides.api.models.manual_webhook import AccessManualWebhook
from fides.api.models.messaging import MessagingConfig
from fides.api.models.messaging_template import MessagingTemplate
from fides.api.models.openid_provider import OpenIDProvider
from fides.api.models.policy import Policy, Rule, RuleTarget
from fides.api.models.privacy_center_config import PrivacyCenterConfig
from fides.api.models.privacy_experience import (
    ExperienceConfigTemplate,
    ExperienceTranslation,
    PrivacyExperience,
    PrivacyExperienceConfig,
    PrivacyExperienceConfigHistory,
)
from fides.api.models.privacy_notice import (
    NoticeTranslation,
    PrivacyNotice,
    PrivacyNoticeHistory,
    PrivacyNoticeTemplate,
)
from fides.api.models.privacy_preference import (
    CurrentPrivacyPreference,
    PrivacyPreferenceHistory,
    ServedNoticeHistory,
)
from fides.api.models.privacy_request import PrivacyRequest
from fides.api.models.property import (
    MessagingTemplateToProperty,
    PrivacyExperienceConfigProperty,
    Property,
)
from fides.api.models.registration import UserRegistration
from fides.api.models.storage import StorageConfig
from fides.api.models.system_compass_sync import SystemCompassSync
from fides.api.models.system_history import SystemHistory
from fides.api.models.system_manager import SystemManager
from fides.api.models.tcf_purpose_overrides import TCFPurposeOverride



"""add connections to client

Revision ID: d9064e71f69d
Revises: 896ea3803770
Create Date: 2024-08-20 22:11:34.351186

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "d9064e71f69d"
down_revision = "896ea3803770"
branch_labels = None
depends_on = None


def upgrade():
    op.add_column(
        "client",
        sa.Column(
            "connections", sa.ARRAY(sa.String()), server_default="{}", nullable=False
        ),
    )


def downgrade():
    op.drop_column("client", "connections")



"""
Dynamically loads all validation step modules in this directory.

This allows validation steps to be added or removed by simply adding/removing Python files,
without modifying existing code. Each .py file (except __init__.py) should contain a single
validation step.
"""

import importlib
import os
import os.path

# path to the current directory
directory = os.path.dirname(__file__)

# loop through the files in the validation directory
for filename in os.listdir(directory):
    # ignore non-Python files and the __init__.py file
    if filename.endswith(".py") and filename != "__init__.py":
        # import the module
        module = importlib.import_module(f"{__name__}.{filename[:-3]}")



from typing import Dict, List, Optional

from fastapi import Depends, Security
from fastapi.encoders import jsonable_encoder
from fastapi_pagination import Page, Params
from fastapi_pagination.bases import AbstractPage
from fastapi_pagination.ext.sqlalchemy import paginate
from fideslang.validation import FidesKey
from loguru import logger
from pydantic import ValidationError
from sqlalchemy.orm import Session
from starlette.exceptions import HTTPException
from starlette.status import (
    HTTP_200_OK,
    HTTP_204_NO_CONTENT,
    HTTP_400_BAD_REQUEST,
    HTTP_404_NOT_FOUND,
    HTTP_422_UNPROCESSABLE_ENTITY,
    HTTP_500_INTERNAL_SERVER_ERROR,
)

from fides.api.api import deps
from fides.api.common_exceptions import (
    EmailTemplateNotFoundException,
    MessageDispatchException,
    MessagingConfigNotFoundException,
    MessagingTemplateValidationException,
)
from fides.api.models.messaging import (
    MessagingConfig,
    default_messaging_config_key,
    default_messaging_config_name,
    get_messaging_method,
    get_schema_for_secrets,
)
from fides.api.models.messaging_template import DEFAULT_MESSAGING_TEMPLATES
from fides.api.oauth.utils import verify_oauth_client
from fides.api.schemas.api import BulkUpdateFailed
from fides.api.schemas.messaging.messaging import (
    BasicMessagingTemplateRequest,
    BasicMessagingTemplateResponse,
    BulkPutBasicMessagingTemplateResponse,
    MessagingActionType,
    MessagingConfigRequest,
    MessagingConfigRequestBase,
    MessagingConfigResponse,
    MessagingConfigStatus,
    MessagingConfigStatusMessage,
    MessagingMethod,
    MessagingServiceType,
    MessagingTemplateDefault,
    MessagingTemplateWithPropertiesDetail,
    TestMessagingStatusMessage,
    UserEmailInviteStatus,
)
from fides.api.schemas.messaging.messaging_secrets_docs_only import (
    possible_messaging_secrets,
)
from fides.api.schemas.redis_cache import Identity
from fides.api.service.messaging.message_dispatch_service import dispatch_message
from fides.api.service.messaging.messaging_crud_service import (
    create_or_update_basic_templates,
    create_or_update_messaging_config,
    delete_messaging_config,
    delete_template_by_id,
    get_all_basic_messaging_templates,
    get_default_template_by_type,
    get_messaging_config_by_key,
    get_template_by_id,
    update_messaging_config,
)
from fides.api.util.api_router import APIRouter
from fides.api.util.logger import Pii
from fides.common.api.scope_registry import (
    MESSAGING_CREATE_OR_UPDATE,
    MESSAGING_DELETE,
    MESSAGING_READ,
    MESSAGING_TEMPLATE_UPDATE,
)
from fides.common.api.v1.urn_registry import (
    BASIC_MESSAGING_TEMPLATES,
    MESSAGING_ACTIVE_DEFAULT,
    MESSAGING_BY_KEY,
    MESSAGING_CONFIG,
    MESSAGING_DEFAULT,
    MESSAGING_DEFAULT_BY_TYPE,
    MESSAGING_DEFAULT_SECRETS,
    MESSAGING_EMAIL_INVITE_STATUS,
    MESSAGING_SECRETS,
    MESSAGING_STATUS,
    MESSAGING_TEMPLATE_BY_ID,
    MESSAGING_TEMPLATE_DEFAULT_BY_TEMPLATE_TYPE,
    MESSAGING_TEST,
    V1_URL_PREFIX,
)
from fides.config.config_proxy import ConfigProxy

router = APIRouter(tags=["Messaging"], prefix=V1_URL_PREFIX)


@router.post(
    MESSAGING_CONFIG,
    status_code=HTTP_200_OK,
    dependencies=[Security(verify_oauth_client, scopes=[MESSAGING_CREATE_OR_UPDATE])],
    response_model=MessagingConfigResponse,
)
def post_config(
    *,
    db: Session = Depends(deps.get_db),
    messaging_config: MessagingConfigRequest,
) -> MessagingConfigResponse:
    """
    Given a messaging config, create corresponding MessagingConfig object, provided no config already exists
    """

    try:
        return create_or_update_messaging_config(db=db, config=messaging_config)
    except ValueError as e:
        logger.warning(
            "Create failed for messaging config {}: {}",
            messaging_config.key,
            Pii(str(e)),
        )
        raise HTTPException(
            status_code=HTTP_400_BAD_REQUEST,
            detail=f"Config with key {messaging_config.key} failed to be added: {e}",
        )
    except Exception as exc:
        logger.warning(
            "Create failed for messaging config {}: {}",
            messaging_config.key,
            Pii(str(exc)),
        )
        raise HTTPException(
            status_code=HTTP_500_INTERNAL_SERVER_ERROR,
            detail=f"Config with key {messaging_config.key} failed to be added: {exc}",
        )


@router.patch(
    MESSAGING_BY_KEY,
    dependencies=[Security(verify_oauth_client, scopes=[MESSAGING_CREATE_OR_UPDATE])],
    response_model=MessagingConfigResponse,
)
def patch_config_by_key(
    config_key: FidesKey,
    *,
    db: Session = Depends(deps.get_db),
    messaging_config: MessagingConfigRequest,
) -> Optional[MessagingConfigResponse]:
    """
    Updates config for messaging by key, provided config with key can be found.
    """
    try:
        return update_messaging_config(db=db, key=config_key, config=messaging_config)
    except MessagingConfigNotFoundException:
        logger.warning("No messaging config found with key {}", config_key)
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND,
            detail=f"No messaging config found with key {config_key}",
        )
    except Exception as exc:
        logger.warning(
            "Patch failed for messaging config {}: {}",
            messaging_config.key,
            Pii(str(exc)),
        )
        raise HTTPException(
            status_code=HTTP_500_INTERNAL_SERVER_ERROR,
            detail=f"Config with key {messaging_config.key} failed to be added",
        )


# this needs to come before other `/default/{messaging_type}` routes so that `/status`
# isn't picked up as a path param
@router.get(
    MESSAGING_ACTIVE_DEFAULT,
    dependencies=[Security(verify_oauth_client, scopes=[MESSAGING_READ])],
    response_model=MessagingConfigResponse,
)
def get_active_default_config(*, db: Session = Depends(deps.get_db)) -> MessagingConfig:
    """
    Retrieves the active default messaging config.
    """
    logger.debug("Finding active default messaging config")
    try:
        messaging_config = MessagingConfig.get_active_default(db)
    except ValueError:
        raise HTTPException(
            status_code=HTTP_400_BAD_REQUEST,
            detail="Invalid notification_service_type configured.",
        )
    if not messaging_config:
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND,
            detail="No active default messaging config found.",
        )
    return messaging_config


# this needs to come before other `/default/{messaging_type}` routes so that `/status`
# isn't picked up as a path param
@router.get(
    MESSAGING_STATUS,
    dependencies=[Security(verify_oauth_client, scopes=[MESSAGING_READ])],
    response_model=MessagingConfigStatusMessage,
    responses={
        HTTP_200_OK: {
            "content": {
                "application/json": {
                    "example": {
                        "config_status": "configured",
                        "detail": "Active default messaging service of type mailgun is fully configured",
                    }
                }
            }
        }
    },
)
def get_messaging_status(
    *,
    db: Session = Depends(deps.get_db),
    messaging_method: Optional[MessagingMethod] = None,
) -> MessagingConfigStatusMessage:
    """
    Determines the status of the active default messaging config
    """
    logger.info("Determining active default messaging config status")

    # confirm an active default messaging config is present
    messaging_config = MessagingConfig.get_active_default(db)

    if not messaging_config or (
        messaging_method
        and get_messaging_method(messaging_config.service_type.value)  # type: ignore
        != messaging_method
    ):
        detail = "No active default messaging configuration found"
        if messaging_method:
            detail += f" for {messaging_method}"

        return MessagingConfigStatusMessage(
            config_status=MessagingConfigStatus.not_configured,
            detail=detail,
        )

    try:
        details = messaging_config.details
        MessagingConfigRequestBase.validate_details_schema(
            messaging_config.service_type, details
        )
    except Exception as e:
        logger.error(f"Invalid or unpopulated details on {messaging_config.service_type.value} messaging configuration: {Pii(str(e))}")  # type: ignore
        return MessagingConfigStatusMessage(
            config_status=MessagingConfigStatus.not_configured,
            detail=f"Invalid or unpopulated details on {messaging_config.service_type.value} messaging configuration",  # type: ignore
        )

    # confirm secrets are present and pass validation
    secrets = messaging_config.secrets
    if not secrets:
        return MessagingConfigStatusMessage(
            config_status=MessagingConfigStatus.not_configured,
            detail=f"No secrets found for {messaging_config.service_type.value} messaging configuration",  # type: ignore
        )
    try:
        get_schema_for_secrets(
            service_type=messaging_config.service_type,  # type: ignore
            secrets=secrets,
        )
    except (ValueError, KeyError) as e:
        logger.error(f"Invalid secrets found on {messaging_config.service_type.value} messaging configuration: {Pii(str(e))}")  # type: ignore
        return MessagingConfigStatusMessage(
            config_status=MessagingConfigStatus.not_configured,
            detail=f"Invalid secrets found on {messaging_config.service_type.value} messaging configuration",  # type: ignore
        )
    except ValidationError:
        return MessagingConfigStatusMessage(
            config_status=MessagingConfigStatus.not_configured,
            detail=f"Invalid secrets found on {messaging_config.service_type.value} messaging configuration",  # type: ignore
        )

    return MessagingConfigStatusMessage(
        config_status=MessagingConfigStatus.configured,
        detail=f"Active default messaging service of type {messaging_config.service_type.value} is fully configured",  # type: ignore
    )


@router.put(
    MESSAGING_DEFAULT,
    dependencies=[Security(verify_oauth_client, scopes=[MESSAGING_CREATE_OR_UPDATE])],
    response_model=MessagingConfigResponse,
)
def put_default_config(
    *,
    db: Session = Depends(deps.get_db),
    messaging_config: MessagingConfigRequestBase,
) -> Optional[MessagingConfigResponse]:
    """
    Updates default messaging config for given service type.
    """
    logger.info(
        "Starting upsert for default messaging config of type '{}'",
        messaging_config.service_type,
    )
    incoming_data = messaging_config.model_dump(mode="json")
    existing_default = MessagingConfig.get_by_type(db, messaging_config.service_type)
    if existing_default:
        # take the key of the existing default and add that to the incoming data, to ensure we overwrite the same record
        incoming_data["key"] = existing_default.key
        incoming_data["name"] = existing_default.name
    else:
        # set a key and name for our config if we're creating a new default
        incoming_data["name"] = default_messaging_config_name(
            messaging_config.service_type.value
        )
        incoming_data["key"] = default_messaging_config_key(
            messaging_config.service_type.value
        )
    return create_or_update_messaging_config(
        db, MessagingConfigRequest(**incoming_data)
    )


@router.put(
    MESSAGING_DEFAULT_SECRETS,
    status_code=HTTP_200_OK,
    dependencies=[Security(verify_oauth_client, scopes=[MESSAGING_CREATE_OR_UPDATE])],
    response_model=TestMessagingStatusMessage,
)
def put_default_config_secrets(
    service_type: MessagingServiceType,
    *,
    db: Session = Depends(deps.get_db),
    unvalidated_messaging_secrets: possible_messaging_secrets,
) -> TestMessagingStatusMessage:
    messaging_config = MessagingConfig.get_by_type(db, service_type=service_type)
    if not messaging_config:
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND,
            detail=f"No default messaging config found of type '{service_type}'",
        )
    return update_config_secrets(db, messaging_config, unvalidated_messaging_secrets)


@router.put(
    MESSAGING_SECRETS,
    status_code=HTTP_200_OK,
    dependencies=[Security(verify_oauth_client, scopes=[MESSAGING_CREATE_OR_UPDATE])],
    response_model=TestMessagingStatusMessage,
)
def put_config_secrets(
    config_key: FidesKey,
    *,
    db: Session = Depends(deps.get_db),
    unvalidated_messaging_secrets: possible_messaging_secrets,
) -> TestMessagingStatusMessage:
    """
    Add or update secrets for messaging config.
    """
    logger.info("Finding messaging config with key '{}'", config_key)
    messaging_config = MessagingConfig.get_by(db=db, field="key", value=config_key)
    if not messaging_config:
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND,
            detail=f"No messaging configuration with key {config_key}.",
        )
    return update_config_secrets(db, messaging_config, unvalidated_messaging_secrets)


def update_config_secrets(
    db: Session,
    messaging_config: MessagingConfig,
    unvalidated_messaging_secrets: possible_messaging_secrets,
) -> TestMessagingStatusMessage:
    try:
        secrets_schema = get_schema_for_secrets(
            service_type=messaging_config.service_type,  # type: ignore
            secrets=unvalidated_messaging_secrets,
        )
    except KeyError as exc:
        raise HTTPException(
            status_code=HTTP_422_UNPROCESSABLE_ENTITY,
            detail=exc.args[0],
        )
    except ValidationError as exc:
        # Remove url, input, and ctx from response. This is to prevent leaking sensitive information.
        errors = exc.errors(include_url=False, include_input=False)
        for err in errors:
            err.pop("ctx", None)

        raise HTTPException(
            status_code=HTTP_422_UNPROCESSABLE_ENTITY,
            detail=jsonable_encoder(errors),
        )

    logger.info(
        "Updating messaging config secrets for config with key '{}'",
        messaging_config.key,
    )
    try:
        messaging_config.set_secrets(db=db, messaging_secrets=secrets_schema.model_dump(mode="json"))  # type: ignore
    except ValueError as exc:
        raise HTTPException(
            status_code=HTTP_400_BAD_REQUEST,
            detail=exc.args[0],
        )
    msg = f"Secrets updated for MessagingConfig with key: {messaging_config.key}."
    # todo- implement test status for messaging service
    return TestMessagingStatusMessage(msg=msg, test_status=None)


@router.get(
    MESSAGING_CONFIG,
    dependencies=[Security(verify_oauth_client, scopes=[MESSAGING_READ])],
    response_model=Page[MessagingConfigResponse],
)
def get_configs(
    *, db: Session = Depends(deps.get_db), params: Params = Depends()
) -> AbstractPage[MessagingConfig]:
    """
    Retrieves configs for messaging.
    """
    logger.debug(
        "Finding all messaging configurations with pagination params {}", params
    )
    return paginate(
        MessagingConfig.query(db=db).order_by(MessagingConfig.created_at.desc()),
        params=params,
    )


@router.get(
    MESSAGING_BY_KEY,
    dependencies=[Security(verify_oauth_client, scopes=[MESSAGING_READ])],
    response_model=MessagingConfigResponse,
)
def get_config_by_key(
    config_key: FidesKey, *, db: Session = Depends(deps.get_db)
) -> MessagingConfigResponse:
    """
    Retrieves configs for messaging service by key.
    """
    logger.debug("Finding messaging config with key '{}'", config_key)

    try:
        return get_messaging_config_by_key(db=db, key=config_key)
    except MessagingConfigNotFoundException as e:
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND,
            detail=e.message,
        )


@router.get(
    MESSAGING_DEFAULT_BY_TYPE,
    dependencies=[Security(verify_oauth_client, scopes=[MESSAGING_READ])],
    response_model=MessagingConfigResponse,
)
def get_default_config_by_type(
    service_type: MessagingServiceType, *, db: Session = Depends(deps.get_db)
) -> MessagingConfig:
    """
    Retrieves default config for messaging service by type.
    """
    logger.debug("Finding default messaging config of type '{}'", service_type)

    messaging_config = MessagingConfig.get_by_type(db, service_type)
    if not messaging_config:
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND,
            detail=f"No default messaging config found of type '{service_type}'",
        )
    return messaging_config


@router.delete(
    MESSAGING_BY_KEY,
    status_code=HTTP_204_NO_CONTENT,
    dependencies=[Security(verify_oauth_client, scopes=[MESSAGING_DELETE])],
)
def delete_config_by_key(
    config_key: FidesKey, *, db: Session = Depends(deps.get_db)
) -> None:
    """
    Deletes messaging configs by key.
    """
    try:
        delete_messaging_config(db=db, key=config_key)
    except MessagingConfigNotFoundException as e:
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND,
            detail=e.message,
        )


@router.post(
    MESSAGING_TEST,
    status_code=HTTP_200_OK,
    dependencies=[Security(verify_oauth_client, scopes=[MESSAGING_CREATE_OR_UPDATE])],
    response_model=Dict[str, str],
    responses={
        HTTP_200_OK: {
            "content": {
                "application/json": {
                    "example": {
                        "details": "Test message successfully sent",
                    }
                }
            }
        }
    },
)
def send_test_message(
    message_info: Identity,
    db: Session = Depends(deps.get_db),
    config_proxy: ConfigProxy = Depends(deps.get_config_proxy),
) -> Dict[str, str]:
    """Sends a test message."""
    try:
        dispatch_message(
            db,
            action_type=MessagingActionType.TEST_MESSAGE,
            to_identity=message_info,
            service_type=config_proxy.notifications.notification_service_type,
        )
    except MessageDispatchException as e:
        raise HTTPException(
            status_code=400, detail=f"There was an error sending the test message: {e}"
        )
    return {"details": "Test message successfully sent"}


@router.get(
    BASIC_MESSAGING_TEMPLATES,
    dependencies=[Security(verify_oauth_client, scopes=[MESSAGING_TEMPLATE_UPDATE])],
    response_model=List[BasicMessagingTemplateResponse],
)
def get_basic_messaging_templates(
    *, db: Session = Depends(deps.get_db)
) -> List[BasicMessagingTemplateResponse]:
    """Returns the available messaging templates, augments the models with labels to be used in the UI."""
    return [
        BasicMessagingTemplateResponse(
            type=template.type,
            content=template.content,
            label=DEFAULT_MESSAGING_TEMPLATES.get(template.type, {}).get("label", None),
        )
        for template in get_all_basic_messaging_templates(db=db)
    ]


@router.put(
    BASIC_MESSAGING_TEMPLATES,
    dependencies=[Security(verify_oauth_client, scopes=[MESSAGING_TEMPLATE_UPDATE])],
)
def update_basic_messaging_templates(
    templates: List[BasicMessagingTemplateRequest],
    *,
    db: Session = Depends(deps.get_db),
) -> BulkPutBasicMessagingTemplateResponse:
    """Updates the messaging templates and reverts empty subject or body values to the default values."""

    succeeded = []
    failed = []

    for template in templates:
        template_type = template.type
        content = template.content

        try:
            default_template = DEFAULT_MESSAGING_TEMPLATES.get(template_type)
            if not default_template:
                raise ValueError("Invalid template type.")

            content["subject"] = (
                content["subject"] or default_template["content"]["subject"]
            )
            content["body"] = content["body"] or default_template["content"]["body"]

            # For Basic Messaging Templates, we ignore the is_enabled flag at runtime. This is because
            # enabling/disabling by template is only supported for property-specific messaging templates,
            # not basic templates.
            create_or_update_basic_templates(
                db,
                data={"type": template_type, "content": content, "is_enabled": False},
            )

            succeeded.append(
                BasicMessagingTemplateResponse(
                    type=template_type,
                    content=content,
                    label=default_template.get("label"),
                )
            )

        except ValueError as e:
            failed.append(
                BulkUpdateFailed(message=str(e), data=template.model_dump(mode="json"))
            )
        except Exception:
            failed.append(
                BulkUpdateFailed(
                    message="Unexpected error updating template.",
                    data=template.model_dump(mode="json"),
                )
            )

    return BulkPutBasicMessagingTemplateResponse(succeeded=succeeded, failed=failed)


@router.get(
    MESSAGING_TEMPLATE_DEFAULT_BY_TEMPLATE_TYPE,
    dependencies=[Security(verify_oauth_client, scopes=[MESSAGING_TEMPLATE_UPDATE])],
    response_model=MessagingTemplateDefault,
)
def get_default_messaging_template(
    template_type: MessagingActionType,
) -> MessagingTemplateDefault:
    """
    Retrieves default messaging template by template type.
    """
    logger.debug(
        "Finding default messaging template of template type '{}'", template_type
    )
    try:
        return get_default_template_by_type(template_type)
    except MessagingTemplateValidationException as e:
        raise HTTPException(
            status_code=HTTP_400_BAD_REQUEST,
            detail=e.message,
        )


@router.get(
    MESSAGING_TEMPLATE_BY_ID,
    dependencies=[Security(verify_oauth_client, scopes=[MESSAGING_TEMPLATE_UPDATE])],
    response_model=MessagingTemplateWithPropertiesDetail,
)
def get_messaging_template_by_id(
    template_id: str,
    *,
    db: Session = Depends(deps.get_db),
) -> MessagingTemplateWithPropertiesDetail:
    """
    Retrieves messaging template by template tid.
    """
    logger.debug("Finding messaging template with id '{}'", template_id)

    try:
        messaging_template = get_template_by_id(db, template_id)
        return MessagingTemplateWithPropertiesDetail(
            id=template_id,
            type=messaging_template.type,
            content=messaging_template.content,
            is_enabled=messaging_template.is_enabled,
            properties=messaging_template.properties,
        )
    except EmailTemplateNotFoundException as e:
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND,
            detail=e.message,
        )


@router.delete(
    MESSAGING_TEMPLATE_BY_ID,
    dependencies=[Security(verify_oauth_client, scopes=[MESSAGING_TEMPLATE_UPDATE])],
    status_code=HTTP_204_NO_CONTENT,
    response_model=None,
)
def delete_messaging_template_by_id(
    template_id: str,
    *,
    db: Session = Depends(deps.get_db),
) -> None:
    """
    Deletes messaging template by template id.
    """
    logger.info("Deleting messaging template with id '{}'", template_id)
    try:
        delete_template_by_id(db, template_id)
    except EmailTemplateNotFoundException as e:
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND,
            detail=e.message,
        )
    except MessagingTemplateValidationException as e:
        raise HTTPException(status_code=HTTP_400_BAD_REQUEST, detail=e.message)


@router.get(MESSAGING_EMAIL_INVITE_STATUS)
def user_email_invite_status(
    db: Session = Depends(deps.get_db),
    config_proxy: ConfigProxy = Depends(deps.get_config_proxy),
) -> UserEmailInviteStatus:
    """Returns whether or not all the necessary configurations are in place to be able to invite a user via email."""

    messaging_status = get_messaging_status(
        db=db, messaging_method=MessagingMethod.EMAIL
    )
    return UserEmailInviteStatus(
        enabled=(
            messaging_status.config_status == MessagingConfigStatus.configured
            and config_proxy.admin_ui.url is not None
        )
    )



from typing import Dict, List, Optional, Union

from fideslang.models import DataSubject, FidesModel, Organization, System
from fideslang.parse import parse_dict

from fides.common.utils import echo_green, echo_red, pretty_echo
from fides.core.api_helpers import (
    get_server_resource,
    get_server_resources,
    list_server_resources,
)


def audit_systems(
    url: str,
    headers: Dict[str, str],
    include_keys: Optional[List] = None,
) -> None:
    """
    Audits the system resources from the server for compliance.

    This should be flexible enough to accept further audits in
    the future and return them as well as being called from other
    audit functions.
    """
    # get the server resources for systems
    system_resources: Optional[Union[List[FidesModel], List[Dict]]]

    if include_keys:
        system_resources = get_server_resources(url, "system", include_keys, headers)
    else:
        system_resources = list_server_resources(
            url, headers, "system", exclude_keys=[]
        )

    if not system_resources:
        print("No system resources were found.")
        return

    print(f"Found {len(system_resources)} System resource(s) to audit...")
    audit_findings = 0
    for system in system_resources:
        pretty_echo(
            f"Auditing System: {system.name if isinstance(system, FidesModel) else system['name']}"
        )
        assert isinstance(system, System)
        new_findings = validate_system_attributes(system, url, headers)
        audit_findings = audit_findings + new_findings

    if audit_findings > 0:
        print(
            f"{audit_findings} issue(s) were detected in auditing system completeness."
        )
    else:
        echo_green("All audited system resource(s) compliant!")


def validate_system_attributes(
    system: System,
    url: str,
    headers: Dict[str, str],
) -> int:
    """
    Validates one or multiple attributes are set on a system
    """

    new_findings = 0
    if system.administrating_department == "Not defined":
        echo_red(
            f"{system.name} should have a responsible group, defined as 'administrating_department'."
        )
        new_findings += 1

    for privacy_declaration in system.privacy_declarations:
        for data_subject_fides_key in privacy_declaration.data_subjects:
            raw_data_subject = get_server_resource(
                url, "data_subject", data_subject_fides_key, headers
            )
            data_subject = parse_dict(
                resource_type="data_subject",
                resource=raw_data_subject,
                from_server=True,
            )
            assert isinstance(data_subject, DataSubject)
            data_subject_findings = audit_data_subject_attributes(
                data_subject, system.name or ""
            )
            new_findings += data_subject_findings
    return new_findings


def audit_data_subject_attributes(data_subject: DataSubject, system_name: str) -> int:
    """
    Audits the extended attributes for a DataSubject
    """
    data_subject_list = ["rights", "automated_decisions_or_profiling"]
    findings = 0
    for attribute in data_subject_list:
        attribute_is_set = getattr(data_subject, attribute) is not None
        compliance_messaging(
            attribute_is_set, data_subject.fides_key, attribute, system_name
        )
        if not attribute_is_set:
            findings += 1
    return findings


def compliance_messaging(
    compliant: bool, fides_key: str, attribute: str, resource_name: str
) -> None:
    """
    Prints the positive confirmation message for a data map compliant attribute.
    """
    if compliant:
        print(f"{attribute} for {fides_key} in {resource_name} is compliant")
    else:
        echo_red(f"{fides_key} missing {attribute} in {resource_name}.")


def audit_organizations(
    url: str,
    headers: Dict[str, str],
    include_keys: Optional[List] = None,
) -> None:
    """
    Validates the extra attributes for an Organization are
    correctly populated
    """

    organization_resources: Optional[List[FidesModel]]

    if include_keys:
        organization_resources = get_server_resources(
            url, "organization", include_keys, headers
        )
    else:
        raw_organization_resources = list_server_resources(
            url, headers, "organization", exclude_keys=[]
        )
        if raw_organization_resources:
            organization_resources = [
                # Parse this into a FidesModel here so we know we're
                # always using a FidesModel downstream
                parse_dict(
                    resource=organization,
                    resource_type="organization",
                    from_server=True,
                )
                for organization in raw_organization_resources
            ]
        else:
            organization_resources = []

    if not organization_resources:
        print("No organization resources were found.")
        return

    print(f"Found {len(organization_resources)} Organization resource(s) to audit...")

    audit_findings = 0
    for organization in organization_resources:
        print(f"Auditing Organization: {organization.name or organization.fides_key}")
        assert isinstance(organization, Organization)
        new_findings = audit_organization_attributes(organization)
        audit_findings += new_findings
    if audit_findings > 0:
        print(
            f"{audit_findings} issue(s) were detected in auditing organization completeness."
        )
    else:
        echo_green("All audited organization resource(s) compliant!")


def audit_organization_attributes(organization: Organization) -> int:
    """
    Audits the list of specified attributes is set properly on the Organization.
    returns the count of findings
    """

    organization_attributes = [
        "controller",
        "data_protection_officer",
        "representative",
        "security_policy",
    ]
    audit_findings = 0
    for attribute in organization_attributes:
        attribute_is_set = getattr(organization, attribute) is not None
        compliance_messaging(
            attribute_is_set, organization.fides_key, attribute, organization.name or ""
        )
        if not attribute_is_set:
            audit_findings += 1

    return audit_findings



"""add dynamodb to connector list

Revision ID: fc04e3e637c0
Revises: 15a3e7483249
Create Date: 2023-04-14 10:19:50.681752

"""

from alembic import op

# revision identifiers, used by Alembic.
revision = "fc04e3e637c0"
down_revision = "15a3e7483249"
branch_labels = None
depends_on = None


def upgrade():
    # Add 'dynamodb' to ConnectionType enum
    op.execute("alter type connectiontype rename to connectiontype_old")
    op.execute(
        "create type connectiontype as enum('postgres', 'mongodb', 'mysql', 'https', "
        "'snowflake', 'redshift', 'mssql', 'mariadb', 'bigquery', 'saas', 'manual', "
        "'manual_webhook', 'timescale', 'fides', 'sovrn', 'attentive', 'dynamodb')"
    )
    op.execute(
        (
            "alter table connectionconfig alter column connection_type type connectiontype using "
            "connection_type::text::connectiontype"
        )
    )
    op.execute("drop type connectiontype_old")


def downgrade():
    # Remove dynamodb from the connectiontype enum
    op.execute("delete from connectionconfig where connection_type in ('dynamodb')")
    op.execute("alter type connectiontype rename to connectiontype_old")
    op.execute(
        "create type connectiontype as enum('postgres', 'mongodb', 'mysql', 'https', "
        "'snowflake', 'redshift', 'mssql', 'mariadb', 'bigquery', 'saas', 'manual', "
        "'email', 'manual_webhook', 'timescale', 'fides', 'sovrn', 'attentive')"
    )
    op.execute(
        (
            "alter table connectionconfig alter column connection_type type connectiontype using "
            "connection_type::text::connectiontype"
        )
    )
    op.execute("drop type connectiontype_old")



"""update discovery monitor config model frequency and databases

Revision ID: ad0109b041b3
Revises: fc2b2c06e595
Create Date: 2024-05-13 14:13:01.956663

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "ad0109b041b3"
down_revision = "fc2b2c06e595"
branch_labels = None
depends_on = None


def upgrade():
    op.add_column(
        "monitorconfig",
        sa.Column(
            "databases", sa.ARRAY(sa.String()), server_default="{}", nullable=False
        ),
    )
    op.add_column(
        "monitorconfig",
        sa.Column(
            "monitor_execution_trigger",
            postgresql.JSONB(astext_type=sa.Text()),
            nullable=True,
        ),
    )


def downgrade():
    op.drop_column("monitorconfig", "monitor_execution_trigger")
    op.drop_column("monitorconfig", "databases")



"""track notices served

Revision ID: 7c562441c589
Revises: 7315b9d7fda6
Create Date: 2023-07-10 16:42:37.433995

"""

import sqlalchemy as sa
import sqlalchemy_utils
from alembic import op

# revision identifiers, used by Alembic.
revision = "7c562441c589"
down_revision = "7315b9d7fda6"
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        "servednoticehistory",
        sa.Column("id", sa.String(length=255), nullable=False),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "anonymized_ip_address",
            sqlalchemy_utils.types.encrypted.encrypted_type.StringEncryptedType(),
            nullable=True,
        ),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "email",
            sqlalchemy_utils.types.encrypted.encrypted_type.StringEncryptedType(),
            nullable=True,
        ),
        sa.Column(
            "fides_user_device",
            sqlalchemy_utils.types.encrypted.encrypted_type.StringEncryptedType(),
            nullable=True,
        ),
        sa.Column("hashed_email", sa.String(), nullable=True),
        sa.Column("hashed_fides_user_device", sa.String(), nullable=True),
        sa.Column("hashed_phone_number", sa.String(), nullable=True),
        sa.Column(
            "phone_number",
            sqlalchemy_utils.types.encrypted.encrypted_type.StringEncryptedType(),
            nullable=True,
        ),
        sa.Column("request_origin", sa.String(), nullable=True),
        sa.Column("url_recorded", sa.String(), nullable=True),
        sa.Column(
            "user_agent",
            sqlalchemy_utils.types.encrypted.encrypted_type.StringEncryptedType(),
            nullable=True,
        ),
        sa.Column("user_geography", sa.String(), nullable=True),
        sa.Column("acknowledge_mode", sa.Boolean(), nullable=True),
        sa.Column("serving_component", sa.String(), nullable=False),
        sa.Column("fides_user_device_provided_identity_id", sa.String(), nullable=True),
        sa.Column("privacy_experience_config_history_id", sa.String(), nullable=True),
        sa.Column("privacy_experience_id", sa.String(), nullable=True),
        sa.Column("privacy_notice_history_id", sa.String(), nullable=False),
        sa.Column("provided_identity_id", sa.String(), nullable=True),
        sa.ForeignKeyConstraint(
            ["fides_user_device_provided_identity_id"],
            ["providedidentity.id"],
        ),
        sa.ForeignKeyConstraint(
            ["privacy_experience_config_history_id"],
            ["privacyexperienceconfighistory.id"],
        ),
        sa.ForeignKeyConstraint(
            ["privacy_experience_id"],
            ["privacyexperience.id"],
        ),
        sa.ForeignKeyConstraint(
            ["privacy_notice_history_id"],
            ["privacynoticehistory.id"],
        ),
        sa.ForeignKeyConstraint(
            ["provided_identity_id"],
            ["providedidentity.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(
        op.f("ix_servednoticehistory_created_at"),
        "servednoticehistory",
        ["created_at"],
        unique=False,
    )
    op.create_index(
        op.f("ix_servednoticehistory_fides_user_device_provided_identity_id"),
        "servednoticehistory",
        ["fides_user_device_provided_identity_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_servednoticehistory_hashed_email"),
        "servednoticehistory",
        ["hashed_email"],
        unique=False,
    )
    op.create_index(
        op.f("ix_servednoticehistory_hashed_fides_user_device"),
        "servednoticehistory",
        ["hashed_fides_user_device"],
        unique=False,
    )
    op.create_index(
        op.f("ix_servednoticehistory_hashed_phone_number"),
        "servednoticehistory",
        ["hashed_phone_number"],
        unique=False,
    )
    op.create_index(
        op.f("ix_servednoticehistory_id"), "servednoticehistory", ["id"], unique=False
    )
    op.create_index(
        op.f("ix_servednoticehistory_privacy_experience_config_history_id"),
        "servednoticehistory",
        ["privacy_experience_config_history_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_servednoticehistory_privacy_experience_id"),
        "servednoticehistory",
        ["privacy_experience_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_servednoticehistory_privacy_notice_history_id"),
        "servednoticehistory",
        ["privacy_notice_history_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_servednoticehistory_provided_identity_id"),
        "servednoticehistory",
        ["provided_identity_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_servednoticehistory_serving_component"),
        "servednoticehistory",
        ["serving_component"],
        unique=False,
    )
    op.create_index(
        op.f("ix_servednoticehistory_user_geography"),
        "servednoticehistory",
        ["user_geography"],
        unique=False,
    )
    op.create_table(
        "lastservednotice",
        sa.Column("id", sa.String(length=255), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("served_notice_history_id", sa.String(), nullable=False),
        sa.Column("provided_identity_id", sa.String(), nullable=True),
        sa.Column("fides_user_device_provided_identity_id", sa.String(), nullable=True),
        sa.Column("privacy_notice_id", sa.String(), nullable=False),
        sa.Column("privacy_notice_history_id", sa.String(), nullable=False),
        sa.ForeignKeyConstraint(
            ["fides_user_device_provided_identity_id"],
            ["providedidentity.id"],
        ),
        sa.ForeignKeyConstraint(
            ["privacy_notice_history_id"],
            ["privacynoticehistory.id"],
        ),
        sa.ForeignKeyConstraint(
            ["privacy_notice_id"],
            ["privacynotice.id"],
        ),
        sa.ForeignKeyConstraint(
            ["provided_identity_id"],
            ["providedidentity.id"],
        ),
        sa.ForeignKeyConstraint(
            ["served_notice_history_id"],
            ["servednoticehistory.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint(
            "fides_user_device_provided_identity_id",
            "privacy_notice_id",
            name="last_served_fides_user_device_identity_privacy_notice",
        ),
        sa.UniqueConstraint(
            "provided_identity_id",
            "privacy_notice_id",
            name="last_served_identity_privacy_notice",
        ),
    )
    op.create_index(
        op.f("ix_lastservednotice_created_at"),
        "lastservednotice",
        ["created_at"],
        unique=False,
    )
    op.create_index(
        op.f("ix_lastservednotice_fides_user_device_provided_identity_id"),
        "lastservednotice",
        ["fides_user_device_provided_identity_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_lastservednotice_id"), "lastservednotice", ["id"], unique=False
    )
    op.create_index(
        op.f("ix_lastservednotice_privacy_notice_history_id"),
        "lastservednotice",
        ["privacy_notice_history_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_lastservednotice_privacy_notice_id"),
        "lastservednotice",
        ["privacy_notice_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_lastservednotice_provided_identity_id"),
        "lastservednotice",
        ["provided_identity_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_lastservednotice_served_notice_history_id"),
        "lastservednotice",
        ["served_notice_history_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_lastservednotice_updated_at"),
        "lastservednotice",
        ["updated_at"],
        unique=False,
    )

    op.add_column(
        "privacypreferencehistory",
        sa.Column("served_notice_history_id", sa.String(), nullable=True),
    )
    op.create_index(
        op.f("ix_privacypreferencehistory_served_notice_history_id"),
        "privacypreferencehistory",
        ["served_notice_history_id"],
        unique=False,
    )
    op.create_foreign_key(
        "privacy_preference_served_notice_fk",
        "privacypreferencehistory",
        "servednoticehistory",
        ["served_notice_history_id"],
        ["id"],
    )


def downgrade():
    op.drop_constraint(
        "privacy_preference_served_notice_fk",
        "privacypreferencehistory",
        type_="foreignkey",
    )
    op.drop_index(
        op.f("ix_privacypreferencehistory_served_notice_history_id"),
        table_name="privacypreferencehistory",
    )
    op.drop_column("privacypreferencehistory", "served_notice_history_id")
    op.drop_index(op.f("ix_lastservednotice_updated_at"), table_name="lastservednotice")
    op.drop_index(
        op.f("ix_lastservednotice_served_notice_history_id"),
        table_name="lastservednotice",
    )
    op.drop_index(
        op.f("ix_lastservednotice_provided_identity_id"), table_name="lastservednotice"
    )
    op.drop_index(
        op.f("ix_lastservednotice_privacy_notice_id"), table_name="lastservednotice"
    )
    op.drop_index(
        op.f("ix_lastservednotice_privacy_notice_history_id"),
        table_name="lastservednotice",
    )
    op.drop_index(op.f("ix_lastservednotice_id"), table_name="lastservednotice")
    op.drop_index(
        op.f("ix_lastservednotice_fides_user_device_provided_identity_id"),
        table_name="lastservednotice",
    )
    op.drop_index(op.f("ix_lastservednotice_created_at"), table_name="lastservednotice")
    op.drop_table("lastservednotice")
    op.drop_index(
        op.f("ix_servednoticehistory_user_geography"), table_name="servednoticehistory"
    )
    op.drop_index(
        op.f("ix_servednoticehistory_serving_component"),
        table_name="servednoticehistory",
    )
    op.drop_index(
        op.f("ix_servednoticehistory_provided_identity_id"),
        table_name="servednoticehistory",
    )
    op.drop_index(
        op.f("ix_servednoticehistory_privacy_notice_history_id"),
        table_name="servednoticehistory",
    )
    op.drop_index(
        op.f("ix_servednoticehistory_privacy_experience_id"),
        table_name="servednoticehistory",
    )
    op.drop_index(
        op.f("ix_servednoticehistory_privacy_experience_config_history_id"),
        table_name="servednoticehistory",
    )
    op.drop_index(op.f("ix_servednoticehistory_id"), table_name="servednoticehistory")
    op.drop_index(
        op.f("ix_servednoticehistory_hashed_phone_number"),
        table_name="servednoticehistory",
    )
    op.drop_index(
        op.f("ix_servednoticehistory_hashed_fides_user_device"),
        table_name="servednoticehistory",
    )
    op.drop_index(
        op.f("ix_servednoticehistory_hashed_email"), table_name="servednoticehistory"
    )
    op.drop_index(
        op.f("ix_servednoticehistory_fides_user_device_provided_identity_id"),
        table_name="servednoticehistory",
    )
    op.drop_index(
        op.f("ix_servednoticehistory_created_at"), table_name="servednoticehistory"
    )
    op.drop_table("servednoticehistory")



"""add redshift and snowflake support

Revision ID: f206d4e7574d
Revises: 0210948a8147
Create Date: 2021-11-09 20:41:39.287326

"""

from alembic import op

# revision identifiers, used by Alembic.
revision = "f206d4e7574d"
down_revision = "0210948a8147"
branch_labels = None
depends_on = None


def upgrade():
    op.execute("alter type connectiontype rename to connectiontype_old")
    op.execute(
        "create type connectiontype as enum('postgres', 'mongodb', 'mysql', 'https', 'redshift', 'snowflake')"
    )
    op.execute(
        (
            "alter table connectionconfig alter column connection_type type connectiontype using "
            "connection_type::text::connectiontype"
        )
    )
    op.execute("drop type connectiontype_old")


def downgrade():
    op.execute(
        "delete from connectionconfig where connection_type in ('snowflake', 'redshift')"
    )
    op.execute("alter type connectiontype rename to connectiontype_old")
    op.execute(
        "create type connectiontype as enum('postgres', 'mongodb', 'mysql', 'https')"
    )
    op.execute(
        (
            "alter table connectionconfig alter column connection_type type connectiontype using "
            "connection_type::text::connectiontype"
        )
    )
    op.execute("drop type connectiontype_old")



"""create custom_connector_template

Revision ID: 8615ac51e78b
Revises: 39a23bc016ea
Create Date: 2023-03-20 15:11:52.634508

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "8615ac51e78b"
down_revision = "39a23bc016ea"
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        "custom_connector_template",
        sa.Column("id", sa.String(length=255), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("key", sa.String(), nullable=False),
        sa.Column("name", sa.String(), nullable=False),
        sa.Column("config", sa.String(), nullable=False),
        sa.Column("dataset", sa.String(), nullable=False),
        sa.Column("icon", sa.String(), nullable=True),
        sa.Column("functions", sa.String(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(
        op.f("ix_custom_connector_template_id"),
        "custom_connector_template",
        ["id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_custom_connector_template_key"),
        "custom_connector_template",
        ["key"],
        unique=True,
    )


def downgrade():
    op.drop_index(
        op.f("ix_custom_connector_template_key"), table_name="custom_connector_template"
    )
    op.drop_index(
        op.f("ix_custom_connector_template_id"), table_name="custom_connector_template"
    )
    op.drop_table("custom_connector_template")



# pylint: disable=R0401, C0302

from __future__ import annotations

import json
from datetime import datetime, timedelta
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Set, Union

from celery.result import AsyncResult
from loguru import logger
from sqlalchemy import Boolean, Column, DateTime, ForeignKey, Integer, String
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.ext.mutable import MutableDict, MutableList
from sqlalchemy.orm import Query, RelationshipProperty, Session, backref, relationship
from sqlalchemy.orm.dynamic import AppenderQuery
from sqlalchemy_utils.types.encrypted.encrypted_type import (
    AesGcmEngine,
    StringEncryptedType,
)

from fides.api.common_exceptions import (
    IdentityVerificationException,
    ManualWebhookFieldsUnset,
    NoCachedManualWebhookEntry,
    PrivacyRequestPaused,
)
from fides.api.cryptography.cryptographic_util import (
    hash_credential_with_salt,
    hash_value_with_salt,
)
from fides.api.cryptography.identity_salt import get_identity_salt
from fides.api.db.base_class import Base  # type: ignore[attr-defined]
from fides.api.db.base_class import JSONTypeOverride
from fides.api.db.util import EnumColumn
from fides.api.graph.config import (
    ROOT_COLLECTION_ADDRESS,
    TERMINATOR_ADDRESS,
    CollectionAddress,
)
from fides.api.migrations.hash_migration_mixin import HashMigrationMixin
from fides.api.models.attachment import Attachment, AttachmentReference
from fides.api.models.audit_log import AuditLog
from fides.api.models.client import ClientDetail
from fides.api.models.comment import Comment, CommentReference
from fides.api.models.fides_user import FidesUser
from fides.api.models.manual_webhook import AccessManualWebhook
from fides.api.models.policy import (
    Policy,
    PolicyPreWebhook,
    WebhookDirection,
    WebhookTypes,
)
from fides.api.models.pre_approval_webhook import (
    PreApprovalWebhook,
    PreApprovalWebhookReply,
)
from fides.api.models.privacy_request.execution_log import (
    COMPLETED_EXECUTION_LOG_STATUSES,
    EXITED_EXECUTION_LOG_STATUSES,
    ExecutionLog,
)
from fides.api.models.privacy_request.provided_identity import ProvidedIdentity
from fides.api.models.privacy_request.request_task import RequestTask
from fides.api.models.privacy_request.webhook import (
    CallbackType,
    SecondPartyRequestFormat,
    generate_request_callback_pre_approval_jwe,
    generate_request_callback_resume_jwe,
)
from fides.api.schemas.drp_privacy_request import DrpPrivacyRequestCreate
from fides.api.schemas.external_https import SecondPartyResponseFormat
from fides.api.schemas.masking.masking_secrets import MaskingSecretCache
from fides.api.schemas.policy import ActionType, CurrentStep
from fides.api.schemas.privacy_request import (
    CheckpointActionRequired,
    ExecutionLogStatus,
    ManualAction,
    PrivacyRequestSource,
    PrivacyRequestStatus,
)
from fides.api.schemas.redis_cache import (
    CustomPrivacyRequestField as CustomPrivacyRequestFieldSchema,
)
from fides.api.schemas.redis_cache import Identity, LabeledIdentity, MultiValue
from fides.api.tasks import celery_app
from fides.api.util.cache import (
    FidesopsRedis,
    get_all_cache_keys_for_privacy_request,
    get_async_task_tracking_cache_key,
    get_cache,
    get_custom_privacy_request_field_cache_key,
    get_drp_request_body_cache_key,
    get_encryption_cache_key,
    get_identity_cache_key,
    get_masking_secret_cache_key,
)
from fides.api.util.collection_util import Row, extract_key_for_address
from fides.api.util.constants import API_DATE_FORMAT
from fides.api.util.custom_json_encoder import CustomJSONEncoder
from fides.api.util.decrypted_identity_automaton import DecryptedIdentityAutomatonMixin
from fides.api.util.identity_verification import IdentityVerificationMixin
from fides.api.util.logger import Pii
from fides.api.util.logger_context_utils import Contextualizable, LoggerContextKeys
from fides.config import CONFIG

if TYPE_CHECKING:
    from fides.api.models.privacy_request.consent import (  # type: ignore[attr-defined]
        Consent,
        ConsentRequest,
    )


class PrivacyRequest(
    IdentityVerificationMixin, DecryptedIdentityAutomatonMixin, Contextualizable, Base
):  # pylint: disable=R0904,too-many-instance-attributes
    """
    The DB ORM model to describe current and historic PrivacyRequests.
    A privacy request is a database record representing the request's
    progression within the Fides system.
    """

    external_id = Column(String, index=True)
    # When the request was dispatched into the Fides pipeline
    started_processing_at = Column(DateTime(timezone=True), nullable=True)
    # When the request finished or errored in the Fides pipeline
    finished_processing_at = Column(DateTime(timezone=True), nullable=True)
    # When the request was created at the origin
    requested_at = Column(DateTime(timezone=True), nullable=True)
    status = Column(
        EnumColumn(PrivacyRequestStatus),
        index=True,
        nullable=False,
    )
    # When the request was approved/denied
    reviewed_at = Column(DateTime(timezone=True), nullable=True)
    # Who approved/denied the request
    reviewed_by = Column(
        String,
        ForeignKey(FidesUser.id_field_path, ondelete="SET NULL"),
        nullable=True,
    )
    submitted_by = Column(
        String,
        ForeignKey(FidesUser.id_field_path, ondelete="SET NULL"),
        nullable=True,
    )
    custom_privacy_request_fields_approved_by = Column(
        String,
        ForeignKey(FidesUser.id_field_path, ondelete="SET NULL"),
        nullable=True,
    )
    client_id = Column(
        String,
        ForeignKey(ClientDetail.id_field_path),
        nullable=True,
    )
    client = relationship(
        ClientDetail,
        backref="privacy_requests",
    )  # Which client submitted the privacy request
    origin = Column(String, nullable=True)  # The origin from the HTTP request
    policy_id = Column(
        String,
        ForeignKey(Policy.id_field_path),
    )
    policy = relationship(
        Policy,
        backref="privacy_requests",
    )
    attachments = relationship(
        Attachment,
        secondary="attachment_reference",
        primaryjoin="PrivacyRequest.id == AttachmentReference.reference_id",
        secondaryjoin="Attachment.id == AttachmentReference.attachment_id",
        order_by="Attachment.created_at",
    )
    comments = relationship(
        Comment,
        secondary="comment_reference",
        primaryjoin="PrivacyRequest.id == CommentReference.reference_id",
        secondaryjoin="Comment.id == CommentReference.comment_id",
        order_by="Comment.created_at",
    )
    property_id = Column(String, nullable=True)

    cancel_reason = Column(String(200))
    canceled_at = Column(DateTime(timezone=True), nullable=True)
    consent_preferences = Column(MutableList.as_mutable(JSONB), nullable=True)
    source = Column(EnumColumn(PrivacyRequestSource), nullable=True)

    # A PrivacyRequest can be soft deleted, so we store when it was deleted
    deleted_at = Column(DateTime(timezone=True), nullable=True)
    # and who deleted it
    deleted_by = Column(String, nullable=True)

    # passive_deletes="all" prevents execution logs from having their privacy_request_id set to null when
    # a privacy_request is deleted.  We want to retain for record-keeping.
    execution_logs = relationship(
        "ExecutionLog",
        backref="privacy_request",
        lazy="dynamic",
        passive_deletes="all",
        primaryjoin="foreign(ExecutionLog.privacy_request_id)==PrivacyRequest.id",
    )

    # passive_deletes="all" prevents audit logs from having their privacy_request_id set to null when
    # a privacy_request is deleted.  We want to retain for record-keeping.
    audit_logs = relationship(
        AuditLog,
        backref="privacy_request",
        lazy="dynamic",
        passive_deletes="all",
        primaryjoin="foreign(AuditLog.privacy_request_id)==PrivacyRequest.id",
    )

    privacy_request_error = relationship(
        "PrivacyRequestError",
        back_populates="privacy_request",
        cascade="delete, delete-orphan",
    )

    reviewer = relationship(
        FidesUser,
        backref=backref("privacy_requests", passive_deletes=True),
        foreign_keys=[reviewed_by],
    )

    pre_approval_webhook_replies = relationship(
        PreApprovalWebhookReply,
        back_populates="privacy_request",
    )

    paused_at = Column(DateTime(timezone=True), nullable=True)
    identity_verified_at = Column(DateTime(timezone=True), nullable=True)
    custom_privacy_request_fields_approved_at = Column(
        DateTime(timezone=True), nullable=True
    )
    due_date = Column(DateTime(timezone=True), nullable=True)
    awaiting_email_send_at = Column(DateTime(timezone=True), nullable=True)

    # Encrypted filtered access results saved for later retrieval
    filtered_final_upload = Column(  # An encrypted JSON String - Dict[Dict[str, List[Row]]] - rule keys mapped to the filtered access results
        StringEncryptedType(
            type_in=JSONTypeOverride,
            key=CONFIG.security.app_encryption_key,
            engine=AesGcmEngine,
            padding="pkcs5",
        ),
    )

    # Encrypted filtered access results saved for later retrieval
    access_result_urls = Column(  # An encrypted JSON String - Dict[Dict[str, List[Row]]] - rule keys mapped to the filtered access results
        StringEncryptedType(
            type_in=JSONTypeOverride,
            key=CONFIG.security.app_encryption_key,
            engine=AesGcmEngine,
            padding="pkcs5",
        ),
    )

    # Non-DB fields that are optionally added throughout the codebase
    action_required_details: Optional[CheckpointActionRequired] = None
    execution_and_audit_logs_by_dataset: Optional[property] = None
    resume_endpoint: Optional[str] = None

    request_tasks: RelationshipProperty[AppenderQuery] = relationship(
        "RequestTask",
        back_populates="privacy_request",
        lazy="dynamic",
        order_by="RequestTask.created_at",
    )

    @property
    def days_left(self: PrivacyRequest) -> Union[int, None]:
        if self.due_date is None:
            return None

        delta = self.due_date.date() - datetime.utcnow().date()
        return delta.days

    @classmethod
    def create(
        cls, db: Session, *, data: Dict[str, Any], check_name: bool = True
    ) -> PrivacyRequest:
        """
        Check whether this object has been passed a `requested_at` value. Default to
        the current datetime if not.
        """
        if data.get("requested_at", None) is None:
            data["requested_at"] = datetime.utcnow()

        policy: Optional[Policy] = Policy.get_by(
            db=db,
            field="id",
            value=data["policy_id"],
        )

        if policy:
            if policy.execution_timeframe:
                requested_at = data["requested_at"]
                if isinstance(requested_at, str):
                    requested_at = datetime.strptime(requested_at, API_DATE_FORMAT)
                data["due_date"] = requested_at + timedelta(
                    days=policy.execution_timeframe
                )

        return super().create(db=db, data=data, check_name=check_name)

    def clear_cached_values(self) -> None:
        """
        Clears all cached values associated with this privacy request from Redis.
        """
        logger.info(f"Clearing cached values for privacy request {self.id}")
        cache: FidesopsRedis = get_cache()
        all_keys = get_all_cache_keys_for_privacy_request(privacy_request_id=self.id)
        for key in all_keys:
            cache.delete(key)

    def delete(self, db: Session) -> None:
        """
        Clean up the cached and persisted data related to this privacy request before
        deleting this object from the database
        """
        self.clear_cached_values()

        for provided_identity in self.provided_identities:  # type: ignore[attr-defined]
            provided_identity.delete(db=db)
        super().delete(db=db)

    def soft_delete(self, db: Session, user_id: Optional[str]) -> None:
        """
        Soft delete the privacy request, marking it as deleted and setting the user who deleted it.
        """
        self.deleted_at = datetime.utcnow()
        self.deleted_by = user_id
        self.save(db)

    def cache_identity(
        self, identity: Union[Identity, Dict[str, LabeledIdentity]]
    ) -> None:
        """Sets the identity's values at their specific locations in the Fides app cache"""
        cache: FidesopsRedis = get_cache()

        if isinstance(identity, dict):
            identity = Identity(**identity)

        identity_dict: Dict[str, Any] = identity.labeled_dict()

        for key, value in identity_dict.items():
            if value is not None:
                cache.set_with_autoexpire(
                    get_identity_cache_key(self.id, key),
                    FidesopsRedis.encode_obj(value),
                )

    def cache_custom_privacy_request_fields(
        self,
        custom_privacy_request_fields: Optional[
            Dict[str, CustomPrivacyRequestFieldSchema]
        ] = None,
    ) -> None:
        """Sets each of the custom privacy request fields values under their own key in the cache"""
        if not custom_privacy_request_fields:
            return

        if not CONFIG.execution.allow_custom_privacy_request_field_collection:
            return

        if CONFIG.execution.allow_custom_privacy_request_fields_in_request_execution:
            cache: FidesopsRedis = get_cache()
            for key, item in custom_privacy_request_fields.items():
                if item is not None:
                    cache.set_with_autoexpire(
                        get_custom_privacy_request_field_cache_key(self.id, key),
                        json.dumps(item.value, cls=CustomJSONEncoder),
                    )
        else:
            logger.info(
                "Custom fields from privacy request {}, but config setting 'CONFIG.execution.allow_custom_privacy_request_fields_in_request_execution' is set to false and prevents their usage.",
                self.id,
            )

    def persist_identity(self, db: Session, identity: Identity) -> None:
        """
        Stores the identity provided with the privacy request in a secure way, compatible with
        blind indexing for later searching and audit purposes.
        """

        if isinstance(identity, dict):
            identity = Identity(**identity)

        identity_dict = identity.labeled_dict()
        for key, value in identity_dict.items():
            if value is not None:
                if isinstance(value, dict):
                    if "label" in value and "value" in value:
                        label = value["label"]
                        value = value["value"]
                    else:
                        raise RuntimeError(
                            f"Programming error: unexpected dict value '{value}' found in an Identity's `labeled_dict()`!"
                        )
                else:
                    label = None

                hashed_value = ProvidedIdentity.hash_value(value)
                provided_identity_data = {
                    "privacy_request_id": self.id,
                    "field_name": key,
                    # We don't need to manually encrypt this field, it's done at the ORM level
                    "encrypted_value": {"value": value},
                    "hashed_value": hashed_value,
                }

                if label is not None:
                    provided_identity_data["field_label"] = label

                ProvidedIdentity.create(
                    db=db,
                    data=provided_identity_data,
                )

        # Simultaneously add identities to automaton for fuzzy search
        if CONFIG.execution.fuzzy_search_enabled:
            try:
                self.add_identities_to_automaton()
            except Exception as exc:
                # This should never affect the ability to create privacy requests
                logger.error(f"Could not add identities to Automaton: {Pii(str(exc))}")

    def persist_custom_privacy_request_fields(
        self,
        db: Session,
        custom_privacy_request_fields: Dict[str, CustomPrivacyRequestFieldSchema],
    ) -> None:
        if not custom_privacy_request_fields:
            return

        if CONFIG.execution.allow_custom_privacy_request_field_collection:
            for key, item in custom_privacy_request_fields.items():
                if item.value:
                    hashed_value = CustomPrivacyRequestField.hash_value(item.value)
                    CustomPrivacyRequestField.create(
                        db=db,
                        data={
                            "privacy_request_id": self.id,
                            "field_name": key,
                            "field_label": item.label,
                            "encrypted_value": {"value": item.value},
                            "hashed_value": hashed_value,
                        },
                    )
        else:
            logger.info(
                "Custom fields provided in privacy request {}, but config setting 'CONFIG.execution.allow_custom_privacy_request_field_collection' prevents their storage.",
                self.id,
            )

    def get_persisted_identity(self) -> Identity:
        """
        Retrieves persisted identity fields from the DB.
        """
        schema_dict = {}
        for field in self.provided_identities:  # type: ignore[attr-defined]
            value = field.encrypted_value.get("value")
            if field.field_label:
                value = LabeledIdentity(label=field.field_label, value=value)
            schema_dict[field.field_name] = value
        return Identity(**schema_dict)

    def get_persisted_custom_privacy_request_fields(self) -> Dict[str, Any]:
        return {
            field.field_name: {
                "label": field.field_label,
                "value": field.encrypted_value["value"],
            }
            for field in self.custom_fields  # type: ignore[attr-defined]
        }

    def verify_identity(self, db: Session, provided_code: str) -> "PrivacyRequest":
        """Verify the identification code supplied by the user
        If verified, change the status of the request to "pending", and set the datetime the identity was verified.
        """
        if not self.status == PrivacyRequestStatus.identity_unverified:
            raise IdentityVerificationException(
                f"Invalid identity verification request. Privacy request '{self.id}' status = {self.status.value}."  # type: ignore # pylint: disable=no-member
            )

        self._verify_identity(provided_code=provided_code)

        self.status = PrivacyRequestStatus.pending
        self.identity_verified_at = datetime.utcnow()
        self.save(db)
        return self

    def get_cached_encryption_key(self) -> Optional[str]:
        """Gets the cached encryption key for this privacy request."""
        cache: FidesopsRedis = get_cache()
        encryption_key = cache.get(get_encryption_cache_key(self.id, "key"))
        return encryption_key

    def get_cached_task_id(self) -> Optional[str]:
        """Gets the cached task ID for this privacy request."""
        cache: FidesopsRedis = get_cache()
        task_id = cache.get(get_async_task_tracking_cache_key(self.id))
        return task_id

    def get_async_execution_task(self) -> Optional[AsyncResult]:
        """Returns a task reflecting the state of this privacy request's asynchronous execution."""
        task_id = self.get_cached_task_id()
        res: AsyncResult = AsyncResult(task_id)
        return res

    def cache_drp_request_body(self, drp_request_body: DrpPrivacyRequestCreate) -> None:
        """Sets the identity's values at their specific locations in the Fides app cache"""
        cache: FidesopsRedis = get_cache()
        drp_request_body_dict: Dict[str, Any] = dict(drp_request_body)
        for key, value in drp_request_body_dict.items():
            if value is not None:
                # handle nested dict/objects
                if not isinstance(value, (bytes, str, int, float)):
                    cache.set_with_autoexpire(
                        get_drp_request_body_cache_key(self.id, key),
                        repr(value),
                    )
                else:
                    cache.set_with_autoexpire(
                        get_drp_request_body_cache_key(self.id, key),
                        value,
                    )

    def cache_encryption(self, encryption_key: Optional[str] = None) -> None:
        """Sets the encryption key in the Fides app cache if provided"""
        if not encryption_key:
            return

        cache: FidesopsRedis = get_cache()
        cache.set_with_autoexpire(
            get_encryption_cache_key(self.id, "key"),
            encryption_key,
        )

    def cache_masking_secret(self, masking_secret: MaskingSecretCache) -> None:
        """Sets masking encryption secrets in the Fides app cache if provided"""
        if not masking_secret:
            return
        cache: FidesopsRedis = get_cache()
        cache.set_with_autoexpire(
            get_masking_secret_cache_key(
                self.id,
                masking_strategy=masking_secret.masking_strategy,
                secret_type=masking_secret.secret_type,
            ),
            FidesopsRedis.encode_obj(masking_secret.secret),
        )

    def get_cached_identity_data(self) -> Dict[str, Any]:
        """Retrieves any identity data pertaining to this request from the cache"""
        prefix = f"id-{self.id}-identity-*"
        cache: FidesopsRedis = get_cache()
        keys = cache.keys(prefix)
        result = {}
        for key in keys:
            value = cache.get(key)
            if value:
                try:
                    # try parsing the value as JSON
                    parsed_value = json.loads(value)
                except json.JSONDecodeError:
                    # if parsing as JSON fails, assume it's a string.
                    # this is purely for backward compatibility: to ensure
                    # that identity data stored pre-2.34.0 in the "old" format
                    # can still be correctly retrieved from the cache.
                    parsed_value = value
                result[key.split("-")[-1]] = parsed_value
        return result

    def get_cached_custom_privacy_request_fields(self) -> Dict[str, Any]:
        """Retrieves any custom fields pertaining to this request from the cache"""
        prefix = f"id-{self.id}-custom-privacy-request-field-*"
        cache: FidesopsRedis = get_cache()
        keys = cache.keys(prefix)
        result = {}
        for key in keys:
            value = cache.get(key)
            if value:
                result[key.split("-")[-1]] = json.loads(value)
        return result

    def get_results(self) -> Dict[str, Any]:
        """Retrieves all cached identity data associated with this Privacy Request
        Just used in testing
        """
        cache: FidesopsRedis = get_cache()
        result_prefix = f"{self.id}__*"
        return cache.get_encoded_objects_by_prefix(result_prefix)

    def cache_email_connector_template_contents(
        self,
        step: CurrentStep,
        collection: CollectionAddress,
        action_needed: List[ManualAction],
    ) -> None:
        """Cache the raw details needed to email to a third party service regarding action they must complete
        on their end for the given collection"""
        cache_action_required(
            cache_key=f"EMAIL_INFORMATION__{self.id}__{step.value}__{collection.dataset}__{collection.collection}",
            step=step,
            collection=collection,
            action_needed=action_needed,
        )

    def get_email_connector_template_contents_by_dataset(
        self, step: CurrentStep, dataset: str
    ) -> List[CheckpointActionRequired]:
        """Retrieve the raw details to populate an email template for collections on a given dataset."""
        cache: FidesopsRedis = get_cache()
        email_contents: Dict[str, Optional[Any]] = cache.get_encoded_objects_by_prefix(
            f"EMAIL_INFORMATION__{self.id}__{step.value}__{dataset}"
        )

        actions: List[CheckpointActionRequired] = []
        for email_content in email_contents.values():
            if email_content:
                actions.append(
                    _parse_cache_to_checkpoint_action_required(email_content)
                )
        return actions

    def cache_paused_collection_details(
        self,
        step: Optional[CurrentStep] = None,
        collection: Optional[CollectionAddress] = None,
        action_needed: Optional[List[ManualAction]] = None,
    ) -> None:
        """
        Cache details about the paused step, paused collection, and any action needed to resume the privacy request.
        """
        cache_action_required(
            cache_key=f"PAUSED_LOCATION__{self.id}",
            step=step,
            collection=collection,
            action_needed=action_needed,
        )

    def get_paused_collection_details(
        self,
    ) -> Optional[CheckpointActionRequired]:
        """Return details about the paused step, paused collection, and any action needed to resume the paused privacy request.

        The paused step lets us know if we should resume privacy request execution from the "access" or the "erasure"
        portion of the privacy request flow, and the collection tells us where we should cache manual input data for later use,
        In other words, this manual data belongs to this collection.
        """
        return get_action_required_details(cached_key=f"EN_PAUSED_LOCATION__{self.id}")

    def cache_failed_checkpoint_details(
        self,
        step: Optional[CurrentStep] = None,
    ) -> None:
        """
        Cache the checkpoint reached in the Privacy Request so it can be resumed from this point in
        case of failure.

        """
        cache_action_required(
            cache_key=f"FAILED_LOCATION__{self.id}",
            step=step,
            collection=None,  # Deprecated for failed checkpoint details
            action_needed=None,
        )

    def get_failed_checkpoint_details(
        self,
    ) -> Optional[CheckpointActionRequired]:
        """Get the latest checkpoint reached in Privacy Request processing so we know where to resume
        in case of failure.
        """
        return get_action_required_details(cached_key=f"EN_FAILED_LOCATION__{self.id}")

    def cache_manual_webhook_access_input(
        self, manual_webhook: AccessManualWebhook, input_data: Optional[Dict[str, Any]]
    ) -> None:
        """Cache manually added data for the given manual webhook.  This is for use by the *manual_webhook* connector,
        which is *NOT* integrated with the graph.

        Dynamically creates a Pydantic model from the manual_webhook to use to validate the input_data
        """
        cache: FidesopsRedis = get_cache()
        parsed_data = manual_webhook.fields_schema.model_validate(input_data)

        cache.set_encoded_object(
            f"WEBHOOK_MANUAL_ACCESS_INPUT__{self.id}__{manual_webhook.id}",
            parsed_data.model_dump(mode="json"),
        )

    def cache_manual_webhook_erasure_input(
        self, manual_webhook: AccessManualWebhook, input_data: Optional[Dict[str, Any]]
    ) -> None:
        """Cache manually added data for the given manual webhook.  This is for use by the *manual_webhook* connector,
        which is *NOT* integrated with the graph.

        Dynamically creates a Pydantic model from the manual_webhook to use to validate the input_data
        """
        cache: FidesopsRedis = get_cache()
        parsed_data = manual_webhook.erasure_fields_schema.model_validate(input_data)

        cache.set_encoded_object(
            f"WEBHOOK_MANUAL_ERASURE_INPUT__{self.id}__{manual_webhook.id}",
            parsed_data.model_dump(mode="json"),
        )

    def get_manual_webhook_access_input_strict(
        self, manual_webhook: AccessManualWebhook
    ) -> Dict[str, Any]:
        """
        Retrieves manual webhook fields saved to the privacy request in strict mode.
        Fails either if extra saved fields are detected (webhook definition had fields removed) or fields were not
        explicitly set (webhook definition had fields added). This mode lets us know if webhooks data needs to be re-uploaded.

        This is for use by the *manual_webhook* connector which is *NOT* integrated with the graph.
        """
        cached_results: Optional[Dict[str, Any]] = _get_manual_access_input_from_cache(
            privacy_request=self, manual_webhook=manual_webhook
        )

        if cached_results:
            data: Dict[str, Any] = manual_webhook.fields_schema.model_validate(
                cached_results
            ).model_dump(exclude_unset=True)
            if set(data.keys()) != set(
                manual_webhook.fields_schema.model_fields.keys()
            ):
                raise ManualWebhookFieldsUnset(
                    f"Fields unset for privacy_request_id '{self.id}' for connection config '{manual_webhook.connection_config.key}'"
                )
            return data
        raise NoCachedManualWebhookEntry(
            f"No data cached for privacy_request_id '{self.id}' for connection config '{manual_webhook.connection_config.key}'"
        )

    def get_manual_webhook_erasure_input_strict(
        self, manual_webhook: AccessManualWebhook
    ) -> Dict[str, Any]:
        """
        Retrieves manual webhook fields saved to the privacy request in strict mode.
        Fails either if extra saved fields are detected (webhook definition had fields removed) or fields were not
        explicitly set (webhook definition had fields added). This mode lets us know if webhooks data needs to be re-uploaded.

        This is for use by the *manual_webhook* connector which is *NOT* integrated with the graph.
        """
        cached_results: Optional[Dict[str, Any]] = _get_manual_erasure_input_from_cache(
            privacy_request=self, manual_webhook=manual_webhook
        )

        if cached_results:
            data: Dict[str, Any] = manual_webhook.erasure_fields_schema.model_validate(
                cached_results
            ).model_dump(exclude_unset=True)
            if set(data.keys()) != set(
                manual_webhook.erasure_fields_schema.model_fields.keys()
            ):
                raise ManualWebhookFieldsUnset(
                    f"Fields unset for privacy_request_id '{self.id}' for connection config '{manual_webhook.connection_config.key}'"
                )
            return data
        raise NoCachedManualWebhookEntry(
            f"No data cached for privacy_request_id '{self.id}' for connection config '{manual_webhook.connection_config.key}'"
        )

    def get_manual_webhook_access_input_non_strict(
        self, manual_webhook: AccessManualWebhook
    ) -> Dict[str, Any]:
        """Retrieves manual webhook fields saved to the privacy request in non-strict mode.
        Returns None for any fields not explicitly set and ignores extra fields.

        This is for use by the *manual_webhook* connector which is *NOT* integrated with the graph.
        """
        cached_results: Optional[Dict[str, Any]] = _get_manual_access_input_from_cache(
            privacy_request=self, manual_webhook=manual_webhook
        )
        if cached_results:
            return manual_webhook.fields_non_strict_schema.model_validate(
                cached_results
            ).model_dump(mode="json")
        return manual_webhook.empty_fields_dict

    def get_manual_webhook_erasure_input_non_strict(
        self, manual_webhook: AccessManualWebhook
    ) -> Dict[str, Any]:
        """Retrieves manual webhook fields saved to the privacy request in non-strict mode.
        Returns None for any fields not explicitly set and ignores extra fields.

        This is for use by the *manual_webhook* connector which is *NOT* integrated with the graph.
        """
        cached_results: Optional[Dict[str, Any]] = _get_manual_erasure_input_from_cache(
            privacy_request=self, manual_webhook=manual_webhook
        )
        if cached_results:
            return manual_webhook.erasure_fields_non_strict_schema.model_validate(
                cached_results
            ).model_dump(mode="json")
        return manual_webhook.empty_fields_dict

    def cache_data_use_map(self, value: Dict[str, Set[str]]) -> None:
        """
        Cache a dict of collections traversed in the privacy request
        mapped to their associated data uses
        """
        cache: FidesopsRedis = get_cache()
        cache.set_encoded_object(f"DATA_USE_MAP__{self.id}", value)

    def get_cached_data_use_map(self) -> Optional[Dict[str, Set[str]]]:
        """
        Fetch the collection -> data use map cached for this privacy request
        """
        cache: FidesopsRedis = get_cache()
        value_dict: Optional[Dict[str, Optional[Dict[str, Set[str]]]]] = (
            cache.get_encoded_objects_by_prefix(f"DATA_USE_MAP__{self.id}")
        )
        return list(value_dict.values())[0] if value_dict else None

    def trigger_pre_approval_webhook(
        self,
        webhook: PreApprovalWebhook,
        policy_action: Optional[ActionType] = None,
    ) -> None:
        """
        Firing pre-approval webhooks allows the privacy request to be automatically approved if all webhooks
        respond with "eligible" to be approved.

        To respond, the service should send a request to one of the reply-to URLs with the reply-to-token.
        """
        # temp fix for circular dependency
        from fides.api.service.connectors import HTTPSConnector, get_connector

        https_connector: HTTPSConnector = get_connector(webhook.connection_config)  # type: ignore
        request_body = SecondPartyRequestFormat(
            privacy_request_id=self.id,
            privacy_request_status=self.status,
            direction=WebhookDirection.two_way,
            callback_type=CallbackType.pre_approval,
            identity=self.get_cached_identity_data(),
            policy_action=policy_action,
        )
        headers = {
            "reply-to-approve": f"/privacy-request/{self.id}/pre-approve/eligible",
            "reply-to-deny": f"/privacy-request/{self.id}/pre-approve/not-eligible",
            "reply-to-token": generate_request_callback_pre_approval_jwe(webhook),  # type: ignore[arg-type]
        }

        logger.info(
            "Calling pre approval webhook '{}' for privacy_request '{}'",
            webhook.key,
            self.id,
        )
        https_connector.execute(  # type: ignore
            request_body.model_dump(mode="json"),
            response_expected=False,
            additional_headers=headers,
        )

    def trigger_policy_webhook(
        self,
        webhook: WebhookTypes,
        policy_action: Optional[ActionType] = None,
    ) -> None:
        """Trigger a request to a single customer-defined policy webhook. Raises an exception if webhook response
        should cause privacy request execution to stop.

        Pre-Execution webhooks send headers to the webhook in case the service needs to send back instructions
        to halt.  To resume, they use send a request to the reply-to URL with the reply-to-token.
        """
        # temp fix for circular dependency
        from fides.api.service.connectors import HTTPSConnector, get_connector

        https_connector: HTTPSConnector = get_connector(webhook.connection_config)  # type: ignore
        request_body = SecondPartyRequestFormat(
            privacy_request_id=self.id,
            privacy_request_status=self.status,
            direction=webhook.direction.value,  # type: ignore
            callback_type=webhook.prefix,
            identity=self.get_cached_identity_data(),
            policy_action=policy_action,
        )

        headers = {}
        is_pre_webhook = isinstance(webhook, PolicyPreWebhook)
        response_expected = webhook.direction == WebhookDirection.two_way
        if is_pre_webhook and response_expected:
            headers = {
                "reply-to": f"/privacy-request/{self.id}/resume",
                "reply-to-token": generate_request_callback_resume_jwe(webhook),  # type: ignore[arg-type]
            }

        logger.info(
            "Calling webhook '{}' for privacy_request '{}'", webhook.key, self.id
        )
        response: Optional[SecondPartyResponseFormat] = https_connector.execute(  # type: ignore
            request_body.model_dump(mode="json"),
            response_expected=response_expected,
            additional_headers=headers,
        )
        if not response:
            return

        response_body = SecondPartyResponseFormat(**response)  # type: ignore

        # Cache any new identities
        if response_body.derived_identity and any(
            [response_body.derived_identity.model_dump(mode="json").values()]
        ):
            logger.info(
                "Updating known identities on privacy request '{}' from webhook '{}'.",
                self.id,
                webhook.key,
            )
            # Don't persist derived identities because they aren't provided directly
            # by the end user
            self.cache_identity(response_body.derived_identity)

        # Pause execution if instructed
        if response_body.halt and is_pre_webhook:
            raise PrivacyRequestPaused(
                f"Halt instruction received on privacy request '{self.id}'."
            )

        return

    def start_processing(self, db: Session) -> None:
        """Dispatches this PrivacyRequest throughout the Fidesops System"""
        if self.started_processing_at is None:
            self.started_processing_at = datetime.utcnow()
        if self.status in [PrivacyRequestStatus.pending, PrivacyRequestStatus.approved]:
            self.status = PrivacyRequestStatus.in_processing
        self.save(db=db)

    def pause_processing(self, db: Session) -> None:
        """Mark privacy request as paused, and save paused_at"""
        self.update(
            db,
            data={
                "status": PrivacyRequestStatus.paused,
                "paused_at": datetime.utcnow(),
            },
        )

    def pause_processing_for_email_send(self, db: Session) -> None:
        """Put the privacy request in a state of awaiting_email_send"""
        if self.awaiting_email_send_at is None:
            self.awaiting_email_send_at = datetime.utcnow()
        self.status = PrivacyRequestStatus.awaiting_email_send
        self.save(db=db)

    def get_request_task_celery_task_ids(self) -> List[str]:
        """Returns the celery task ids for each of the Request Tasks (subtasks)

        It is possible Request Tasks get queued multiple times, so the celery task
        id returned is the last celery task queued.
        """
        request_task_celery_ids: List[str] = []
        for request_task in self.request_tasks:
            request_task_id: Optional[str] = request_task.get_cached_task_id()
            if request_task_id:
                request_task_celery_ids.append(request_task_id)
        return request_task_celery_ids

    def cancel_processing(self, db: Session, cancel_reason: Optional[str]) -> None:
        """Cancels a privacy request.  Currently should only cancel 'pending' tasks

        Just in case, also tries to cancel sub tasks (Request Tasks) if applicable,
        although these shouldn't exist if the Privacy Request is pending.
        """
        if self.canceled_at is None:
            self.status = PrivacyRequestStatus.canceled
            self.cancel_reason = cancel_reason
            self.canceled_at = datetime.utcnow()
            self.save(db)

            task_ids: List[str] = (
                self.get_request_task_celery_task_ids()
            )  # Celery tasks for sub tasks (DSR 3.0 Request Tasks)
            parent_task_id = (
                self.get_cached_task_id()
            )  # Celery task for current Privacy Request
            if parent_task_id:
                task_ids.append(parent_task_id)

            for celery_task_id in task_ids:
                logger.info("Revoking task {} for request {}", celery_task_id, self.id)
                # Only revokes if execution is not already in progress.
                celery_app.control.revoke(celery_task_id, terminate=False)

    def error_processing(self, db: Session) -> None:
        """Mark privacy request as errored, and note time processing was finished"""
        self.update(
            db,
            data={
                "status": PrivacyRequestStatus.error,
                "finished_processing_at": datetime.utcnow(),
            },
        )

        PrivacyRequestError.create(
            db=db, data={"message_sent": False, "privacy_request_id": self.id}
        )

    def get_log_context(self) -> Dict[LoggerContextKeys, Any]:
        context = {LoggerContextKeys.privacy_request_id: self.id}
        if self.source:
            context[LoggerContextKeys.privacy_request_source] = self.source.value
        return context

    @property
    def access_tasks(self) -> Query:
        """Return existing Access Request Tasks for the current privacy request"""
        return self.request_tasks.filter(RequestTask.action_type == ActionType.access)

    @property
    def erasure_tasks(self) -> Query:
        """Return existing Erasure Request Tasks for the current privacy request"""
        return self.request_tasks.filter(RequestTask.action_type == ActionType.erasure)

    @property
    def consent_tasks(self) -> Query:
        """Return existing Consent Request Tasks for the current privacy request"""
        return self.request_tasks.filter(RequestTask.action_type == ActionType.consent)

    def get_comment_by_id(self, db: Session, comment_id: str) -> Optional[Comment]:
        """Get the comment associated with the privacy request"""
        comment = (
            db.query(Comment)
            .join(CommentReference, Comment.id == CommentReference.comment_id)
            .filter(
                CommentReference.reference_id
                == self.id,  # Ensure the comment is linked to this privacy request
                Comment.id == comment_id,  # Match the specific comment ID
            )
            .first()
        )
        if not comment:
            logger.info(
                f"Comment with id {comment_id} not found on privacy request {self.id}"
            )
        return comment

    def get_attachment_by_id(
        self, db: Session, attachment_id: str
    ) -> Optional[Attachment]:
        """Get the attachment associated with the privacy request"""
        attachment = (
            db.query(Attachment)
            .join(
                AttachmentReference, Attachment.id == AttachmentReference.attachment_id
            )
            .filter(
                AttachmentReference.reference_id == self.id,
                Attachment.id == attachment_id,
            )
            .first()
        )
        if not attachment:
            logger.info(
                f"Attachment with id {attachment_id} not found on privacy request {self.id}"
            )
        return attachment

    def delete_attachment_by_id(self, db: Session, attachment_id: str) -> None:
        """Delete the attachment associated with the privacy request"""
        attachment = self.get_attachment_by_id(db, attachment_id)
        if attachment:
            attachment.delete(db)

    def get_existing_request_task(
        self,
        db: Session,
        action_type: ActionType,
        collection_address: CollectionAddress,
    ) -> Optional[RequestTask]:
        """Returns a Request Task for the current Privacy Request with action type and collection address"""
        return (
            db.query(RequestTask)
            .filter(
                RequestTask.privacy_request_id == self.id,
                RequestTask.action_type == action_type,
                RequestTask.collection_address == collection_address.value,
            )
            .first()
        )

    def get_tasks_by_action(self, action: ActionType) -> Query:
        """Convenience helper to get RequestTasks of a certain action type for the given
        privacy request"""
        if action == ActionType.access:
            return self.access_tasks

        if action == ActionType.erasure:
            return self.erasure_tasks

        if action == ActionType.consent:
            return self.consent_tasks

        raise Exception(f"Unsupported Privacy Request Action Type {action}")

    def get_root_task_by_action(self, action: ActionType) -> RequestTask:
        """Get the root tasks for a specific action"""
        root: Optional[RequestTask] = (
            self.get_tasks_by_action(action)
            .filter(RequestTask.collection_address == ROOT_COLLECTION_ADDRESS.value)
            .first()
        )
        if not root:
            raise Exception(
                f"Expected {action.value.capitalize()} root node cannot be found on privacy request {self.id} "
            )
        assert root  # for mypy
        return root

    def get_terminate_task_by_action(self, action: ActionType) -> RequestTask:
        """Get the terminate task for a specific action"""
        terminate: Optional[RequestTask] = (
            self.get_tasks_by_action(action)
            .filter(RequestTask.collection_address == TERMINATOR_ADDRESS.value)
            .first()
        )
        if not terminate:
            raise Exception(
                f"Expected {action.value.capitalize()} terminate node cannot be found on privacy request {self.id} "
            )
        assert terminate  # for mypy
        return terminate

    def get_raw_access_results(self) -> Dict[str, Optional[List[Row]]]:
        """Retrieve the *raw* access data saved on the individual access nodes

        These shouldn't be returned to the user - they are not filtered by data category
        """
        # For DSR 3.0, pull these off of the RequestTask.access_data fields
        if self.access_tasks.count():
            final_results: Dict = {}
            for task in self.access_tasks.filter(
                RequestTask.status == PrivacyRequestStatus.complete,
                RequestTask.collection_address.notin_(
                    [ROOT_COLLECTION_ADDRESS.value, TERMINATOR_ADDRESS.value]
                ),
            ):
                final_results[task.collection_address] = task.get_access_data()

            return final_results

        # TODO Remove when we stop support for DSR 2.0
        # We will no longer be pulling access results from the cache, but off of Request Tasks instead
        cache: FidesopsRedis = get_cache()
        value_dict = cache.get_encoded_objects_by_prefix(f"{self.id}__access_request")
        # extract request id to return a map of address:value
        number_of_leading_strings_to_exclude = 2
        return {
            extract_key_for_address(k, number_of_leading_strings_to_exclude): v
            for k, v in value_dict.items()
        }

    def get_raw_masking_counts(self) -> Dict[str, int]:
        """For parity, return the rows masked for an erasure request

        This is largely just used for testing
        """
        if self.erasure_tasks.count():
            # For DSR 3.0
            return {
                t.collection_address: t.rows_masked
                for t in self.erasure_tasks.filter(
                    RequestTask.status.in_(COMPLETED_EXECUTION_LOG_STATUSES)
                )
                if not t.is_root_task and not t.is_terminator_task
            }

        # TODO Remove when we stop support for DSR 2.0
        cache: FidesopsRedis = get_cache()
        value_dict = cache.get_encoded_objects_by_prefix(f"{self.id}__erasure_request")
        # extract request id to return a map of address:value
        number_of_leading_strings_to_exclude = 2
        return {extract_key_for_address(k, number_of_leading_strings_to_exclude): v for k, v in value_dict.items()}  # type: ignore

    def get_consent_results(self) -> Dict[str, int]:
        """For parity, return whether a consent request was sent for third
        party consent propagation

        This is largely just used for testing
        """
        if self.consent_tasks.count():
            # For DSR 3.0
            return {
                t.collection_address: t.consent_sent
                for t in self.consent_tasks.filter(
                    RequestTask.status.in_(EXITED_EXECUTION_LOG_STATUSES)
                )
                if not t.is_root_task and not t.is_terminator_task
            }
        # DSR 2.0 does not cache the results so nothing to do here
        return {}

    def save_filtered_access_results(
        self, db: Session, results: Dict[str, Dict[str, List[Row]]]
    ) -> None:
        """
        For access requests, save the access data filtered by data category that we uploaded to the end user

        This is keyed by policy rule key, because we uploaded different packages for different policy rules

        """
        if not self.policy.get_rules_for_action(action_type=ActionType.access):
            return None

        self.filtered_final_upload = results
        self.save(db)

        return None

    def get_filtered_final_upload(self) -> Dict[str, Dict[str, List[Row]]]:
        """Fetched the same filtered access results we uploaded to the user"""
        return self.filtered_final_upload or {}

    def add_success_execution_log(
        self,
        db: Session,
        connection_key: Optional[str],
        dataset_name: Optional[str],
        collection_name: Optional[str],
        message: str,
        action_type: ActionType,
    ) -> ExecutionLog:
        return ExecutionLog.create(
            db=db,
            data={
                "privacy_request_id": self.id,
                "connection_key": connection_key,
                "dataset_name": dataset_name,
                "collection_name": collection_name,
                "status": ExecutionLogStatus.complete,
                "message": message,
                "action_type": action_type,
            },
        )

    def add_skipped_execution_log(
        self,
        db: Session,
        connection_key: Optional[str],
        dataset_name: Optional[str],
        collection_name: Optional[str],
        message: str,
        action_type: ActionType,
    ) -> ExecutionLog:
        return ExecutionLog.create(
            db=db,
            data={
                "privacy_request_id": self.id,
                "connection_key": connection_key,
                "dataset_name": dataset_name,
                "collection_name": collection_name,
                "status": ExecutionLogStatus.skipped,
                "message": message,
                "action_type": action_type,
            },
        )

    def add_error_execution_log(
        self,
        db: Session,
        connection_key: Optional[str],
        dataset_name: Optional[str],
        collection_name: Optional[str],
        message: str,
        action_type: ActionType,
    ) -> ExecutionLog:
        return ExecutionLog.create(
            db=db,
            data={
                "privacy_request_id": self.id,
                "connection_key": connection_key,
                "dataset_name": dataset_name,
                "collection_name": collection_name,
                "status": ExecutionLogStatus.error,
                "message": message,
                "action_type": action_type,
            },
        )


class PrivacyRequestError(Base):
    """The DB ORM model to track PrivacyRequests error message status."""

    message_sent = Column(Boolean, nullable=False, default=False)
    privacy_request_id = Column(
        String,
        ForeignKey(
            PrivacyRequest.id_field_path, ondelete="CASCADE", onupdate="CASCADE"
        ),
        nullable=False,
    )

    privacy_request = relationship(
        PrivacyRequest, back_populates="privacy_request_error"
    )


def _get_manual_access_input_from_cache(
    privacy_request: PrivacyRequest, manual_webhook: AccessManualWebhook
) -> Optional[Dict[str, Any]]:
    """Get raw manual input uploaded to the privacy request for the given webhook
    from the cache without attempting to coerce into a Pydantic schema"""
    cache: FidesopsRedis = get_cache()
    cached_results: Optional[Optional[Dict[str, Any]]] = (
        cache.get_encoded_objects_by_prefix(
            f"WEBHOOK_MANUAL_ACCESS_INPUT__{privacy_request.id}__{manual_webhook.id}"
        )
    )
    if cached_results:
        return list(cached_results.values())[0]
    return None


def _get_manual_erasure_input_from_cache(
    privacy_request: PrivacyRequest, manual_webhook: AccessManualWebhook
) -> Optional[Dict[str, Any]]:
    """Get raw manual input uploaded to the privacy request for the given webhook
    from the cache without attempting to coerce into a Pydantic schema"""
    cache: FidesopsRedis = get_cache()
    cached_results: Optional[Optional[Dict[str, Any]]] = (
        cache.get_encoded_objects_by_prefix(
            f"WEBHOOK_MANUAL_ERASURE_INPUT__{privacy_request.id}__{manual_webhook.id}"
        )
    )
    if cached_results:
        return list(cached_results.values())[0]
    return None


class PrivacyRequestNotifications(Base):
    email = Column(String, nullable=False)
    notify_after_failures = Column(Integer, nullable=False)


class CustomPrivacyRequestField(HashMigrationMixin, Base):
    @declared_attr
    def __tablename__(self) -> str:
        return "custom_privacy_request_field"

    privacy_request_id = Column(
        String,
        ForeignKey(
            PrivacyRequest.id_field_path, ondelete="CASCADE", onupdate="CASCADE"
        ),
    )
    privacy_request = relationship(
        PrivacyRequest,
        backref="custom_fields",
    )

    consent_request_id = Column(String, ForeignKey("consentrequest.id"))
    consent_request = relationship("ConsentRequest", back_populates="custom_fields")

    field_name = Column(
        String,
        index=False,
        nullable=False,
    )
    field_label = Column(
        String,
        index=False,
        nullable=False,
    )
    hashed_value = Column(
        String,
        index=True,
        unique=False,
        nullable=True,
    )  # This field is used as a blind index for exact match searches
    encrypted_value = Column(
        MutableDict.as_mutable(
            StringEncryptedType(
                JSONTypeOverride,
                CONFIG.security.app_encryption_key,
                AesGcmEngine,
                "pkcs5",
            )
        ),
        nullable=True,
    )  # Type bytea in the db

    @classmethod
    def bcrypt_hash_value(
        cls,
        value: MultiValue,
        encoding: str = "UTF-8",
    ) -> Optional[str]:
        """
        Temporary function used to hash values to the previously used bcrypt hashes.
        This can be removed once the bcrypt to SHA-256 migration is complete.
        """

        def hash_single_value(value: Union[str, int]) -> str:
            SALT = "$2b$12$UErimNtlsE6qgYf2BrI1Du"
            value_str = str(value)
            hashed_value = hash_credential_with_salt(
                value_str.encode(encoding),
                SALT.encode(encoding),
            )
            return hashed_value

        if isinstance(value, list):
            # Skip hashing lists: this avoids us hashing and later indexing potentially large values and our index
            # is not useful for array search anyway
            return None
        return hash_single_value(value)

    @classmethod
    def hash_value(
        cls,
        value: MultiValue,
        encoding: str = "UTF-8",
    ) -> Optional[str]:
        """Utility function to hash the value(s) with a generated salt"""

        def hash_single_value(value: Union[str, int]) -> str:
            SALT = get_identity_salt()
            value_str = str(value)
            hashed_value = hash_value_with_salt(
                value_str.encode(encoding),
                SALT.encode(encoding),
            )
            return hashed_value

        if isinstance(value, list):
            # Skip hashing lists: this avoids us hashing and later indexing potentially large values and our index
            # is not useful for array search anyway
            return None
        return hash_single_value(value)

    def migrate_hashed_fields(self) -> None:
        if value := self.encrypted_value.get("value"):
            self.hashed_value = self.hash_value(value)  # type: ignore
        self.is_hash_migrated = True


def cache_action_required(
    cache_key: str,
    step: Optional[CurrentStep] = None,
    collection: Optional[CollectionAddress] = None,
    action_needed: Optional[List[ManualAction]] = None,
) -> None:
    """Generic method to cache information about additional action required for a collection.

    For example, we might pause a privacy request at the access step of the postgres_example:address collection.  The
    user might need to retrieve an "email" field and an "address" field where the customer_id is 22 to resume the request.

    The "step" describes whether action is needed in the access or the erasure portion of the request.
    """
    cache: FidesopsRedis = get_cache()
    action_required: Optional[CheckpointActionRequired] = None
    if step:
        action_required = CheckpointActionRequired(
            step=step, collection=collection, action_needed=action_needed
        )

    cache.set_encoded_object(
        cache_key,
        action_required.model_dump() if action_required else None,
    )


def get_action_required_details(
    cached_key: str,
) -> Optional[CheckpointActionRequired]:
    """Get details about the action required for a given collection.

    The "step" lets us know if action is needed in the "access" or the "erasure" portion of the privacy request flow.
    The "collection" is the node in question, and the "action_needed" describes actions that must be manually
    performed to complete the request.
    """
    cache: FidesopsRedis = get_cache()
    cached_stopped: Optional[dict[str, Any]] = cache.get_encoded_by_key(cached_key)
    if cached_stopped:
        return _parse_cache_to_checkpoint_action_required(cached_stopped)

    return None


def _parse_cache_to_checkpoint_action_required(
    cache: dict[str, Any]
) -> CheckpointActionRequired:
    collection = (
        CollectionAddress(
            cache["collection"]["dataset"],
            cache["collection"]["collection"],
        )
        if cache.get("collection")
        else None
    )
    action_needed = (
        [ManualAction(**action) for action in cache["action_needed"]]
        if cache.get("action_needed")
        else None
    )
    return CheckpointActionRequired(
        step=cache["step"],
        collection=collection,
        action_needed=action_needed,
    )



from fides.api.common_exceptions import ValidationError
from fides.api.schemas.namespace_meta.namespace_meta import NamespaceMeta
from fides.service.dataset.dataset_validator import (
    DatasetValidationContext,
    DatasetValidationStep,
)


class NamespaceMetaValidationStep(DatasetValidationStep):
    """Validates that datasets have required namespace metadata based on connection type"""

    def validate(self, context: DatasetValidationContext) -> None:
        if not context.connection_config:
            return

        connection_type: str = context.connection_config.connection_type.value  # type: ignore[attr-defined]
        namespace_meta_class = NamespaceMeta.get_implementation(connection_type)
        if not namespace_meta_class:
            return

        # Check if dataset has namespace metadata
        namespace_meta = (
            context.dataset.fides_meta.namespace if context.dataset.fides_meta else None
        )

        # Get required secret fields from the namespace meta class
        required_fields = namespace_meta_class.get_fallback_secret_fields()
        has_connection_defaults = all(
            field_name in context.connection_config.secrets
            and context.connection_config.secrets[field_name]
            for field_name, _ in required_fields
        )

        if not namespace_meta and not has_connection_defaults:
            field_descriptions = ", ".join(
                field_label for _, field_label in required_fields
            )
            raise ValidationError(
                f"Dataset for {connection_type} connection must either have namespace metadata "
                f"or the connection must have values for the following fields: {field_descriptions}"
            )
        if namespace_meta:
            try:
                namespace_meta_class(**namespace_meta)
            except Exception as e:
                raise ValidationError(
                    f"Invalid namespace metadata for {connection_type}: {str(e)}"
                )



"""pre_approval_webhook_cleanup

Revision ID: 9e83545ed9b6
Revises: c85a641cc92c
Create Date: 2024-04-26 23:47:27.821607

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "9e83545ed9b6"
down_revision = "c85a641cc92c"
branch_labels = None
depends_on = None


def upgrade():
    # Index PreApproval Webhook Key
    op.create_index(
        op.f("ix_preapprovalwebhook_key"), "preapprovalwebhook", ["key"], unique=True
    )
    # Create Unique Constraint on Preapproval Webhook Name
    op.create_unique_constraint(
        "preapprovalwebhook_name_key", "preapprovalwebhook", ["name"]
    )
    # Remove PreApprovalWebhook.connection_config_id FK
    op.drop_constraint(
        "preapprovalwebhook_connection_config_id_fkey",
        "preapprovalwebhook",
        type_="foreignkey",
    )
    # Restore PreApprovalWebhook.connection_config_id FK with cascade delete
    op.create_foreign_key(
        "preapprovalwebhook_connection_config_id_fkey",
        "preapprovalwebhook",
        "connectionconfig",
        ["connection_config_id"],
        ["id"],
        ondelete="CASCADE",
    )
    # Set Preapproval Webhook Reply to nullable in case the Webhook is deleted
    op.alter_column(
        "preapprovalwebhookreply",
        "webhook_id",
        existing_type=sa.VARCHAR(length=255),
        nullable=True,
    )
    # Index the PreApprovalWebhookReply.privacy_request_id
    op.create_index(
        op.f("ix_preapprovalwebhookreply_privacy_request_id"),
        "preapprovalwebhookreply",
        ["privacy_request_id"],
        unique=False,
    )
    # Index the PreApprovalWebhookReply.webhook_id
    op.create_index(
        op.f("ix_preapprovalwebhookreply_webhook_id"),
        "preapprovalwebhookreply",
        ["webhook_id"],
        unique=False,
    )
    # Remove FK on preapprovalwebhookreply.privacy_request_id
    op.drop_constraint(
        "preapprovalwebhookreply_privacy_request_id_fkey",
        "preapprovalwebhookreply",
        type_="foreignkey",
    )
    # Add FK on preapprovalwebhookreply.webhook_id with ondelete set null
    op.create_foreign_key(
        "preapprovalwebhookreply_webhook_id_fkey",
        "preapprovalwebhookreply",
        "preapprovalwebhook",
        ["webhook_id"],
        ["id"],
        ondelete="SET NULL",
    )
    # Restore FK on preapprovalwebhookreply.privacy_request_id with ondelete set null
    op.create_foreign_key(
        "preapprovalwebhookreply_privacy_request_id_fkey",
        "preapprovalwebhookreply",
        "privacyrequest",
        ["privacy_request_id"],
        ["id"],
        ondelete="SET NULL",
    )


def downgrade():
    op.drop_constraint(
        "preapprovalwebhookreply_privacy_request_id_fkey",
        "preapprovalwebhookreply",
        type_="foreignkey",
    )
    op.drop_constraint(
        "preapprovalwebhookreply_webhook_id_fkey",
        "preapprovalwebhookreply",
        type_="foreignkey",
    )
    op.create_foreign_key(
        "preapprovalwebhookreply_privacy_request_id_fkey",
        "preapprovalwebhookreply",
        "privacyrequest",
        ["privacy_request_id"],
        ["id"],
    )
    op.drop_index(
        op.f("ix_preapprovalwebhookreply_webhook_id"),
        table_name="preapprovalwebhookreply",
    )
    op.drop_index(
        op.f("ix_preapprovalwebhookreply_privacy_request_id"),
        table_name="preapprovalwebhookreply",
    )
    op.drop_constraint(
        "preapprovalwebhook_connection_config_id_fkey",
        "preapprovalwebhook",
        type_="foreignkey",
    )
    op.create_foreign_key(
        "preapprovalwebhook_connection_config_id_fkey",
        "preapprovalwebhook",
        "connectionconfig",
        ["connection_config_id"],
        ["id"],
    )
    op.drop_constraint(
        "preapprovalwebhook_name_key", "preapprovalwebhook", type_="unique"
    )
    op.drop_index(op.f("ix_preapprovalwebhook_key"), table_name="preapprovalwebhook")
